from pangtreebuild.affinity_tree import tree
from pangtreebuild.pangenome import graph


def poagraph_to_fasta(poagraph: graph.Poagraph) -> str:
    """Converts poagraph to FASTA format.

    Args:
        poagraph: Poagraph to be converted.

    Returns:
        Fasta formatted string ready to be saved to file.
    """

    fasta_lines = []
    for seq_id, sequence in poagraph.sequences.items():
        if poagraph.get_sequence_nodes_count(seq_id) == 0:
            continue
        sequence = "".join([poagraph.nodes[node_id].get_base()
                            for path in sequence.paths
                            for node_id in path])
        fasta_lines.append(f">{seq_id}")
        fasta_lines.append(sequence)

    return "\n".join(fasta_lines)


def affinity_tree_to_fasta(poagraph: graph.Poagraph,
                           affinity_tree: tree.AffinityTree) -> str:
    """Converts Affinity tree to FASTA formatted string.

    Args:
        poagraph: Poagraph linked to the affinity_tree.
        affinity_tree: Consensus paths from this affinty tree
            will be converted to FASTA.

    Returns:
        Fasta formatted string with consensus paths from given affinty tree.
    """

    fasta_lines = []
    for affinity_node in affinity_tree.nodes:
        if affinity_node.consensus is None or \
           len(affinity_node.consensus) == 0:
            continue
        sequence = "".join([poagraph.nodes[node_id].get_base()
                            for node_id in affinity_node.consensus])
        missing_mincomp_symbol = "?"
        leaf = "" \
               if len(affinity_node.children) > 1 \
               else ",".join([str(seq_id)
                              for seq_id in affinity_node.sequences])
        fasta_lines.append(f">AffinityNode{affinity_node.id_}|"
                           f"leaf_for={leaf}|"
                           f"""mincomp={affinity_node.mincomp
                                      if affinity_node.mincomp is not None
                                      else missing_mincomp_symbol}|"""
                           f"sequences_count={len(affinity_node.sequences)}|"
                           f"children={str(affinity_node.children)}")
        fasta_lines.append(sequence)

    return "\n".join(fasta_lines)



import unittest
from pathlib import Path

from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.pangenome import builder
from pangtreebuild.pangenome.parameters import missings
from pangtreebuild.tools import pathtools


def nid(x): return graph.NodeID(x)


def bid(x): return graph.BlockID(x)


class DAGMaf2PoagraphConstSymbolProviderTests(unittest.TestCase):
    def setUp(self):
        metadata_path = Path(__file__).parent.joinpath("../seq_metadata.csv").resolve()
        self.metadatacsv = msa.MetadataCSV(pathtools.get_file_content_stringio(metadata_path), metadata_path)
        self.maf_files_dir = Path(__file__).parent.joinpath("maf_files_with_gaps").resolve()
        self.missing_n = missings.MissingBase()

    def test_1_missing_sequence_start(self):
        maf_path = self.maf_files_dir.joinpath("test_1_missing_sequence_start.maf")
        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base(self.missing_n.value), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(1), base=graph.Base(self.missing_n.value), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(2), base=graph.Base(self.missing_n.value), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(3), base=graph.Base('A'), aligned_to=nid(4)),
            graph.Node(node_id=nid(4), base=graph.Base('G'), aligned_to=nid(3)),
            graph.Node(node_id=nid(5), base=graph.Base('G'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(6), base=graph.Base('G'), aligned_to=nid(7)),
            graph.Node(node_id=nid(7), base=graph.Base('T'), aligned_to=nid(6)),
            graph.Node(node_id=nid(8), base=graph.Base('C'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(9), base=graph.Base('A'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(10), base=graph.Base('G'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(11), base=graph.Base('T'), aligned_to=None, block_id=bid(0))
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 3, 5, 6, 11])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [4, 5, 7, 8, 9, 10, 11])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            missings.ConstBaseProvider(self.missing_n),
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_2_missing_sequence_end(self):
        maf_path = self.maf_files_dir.joinpath("test_2_missing_sequence_end.maf")

        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=nid(1)),
            graph.Node(node_id=nid(1), base=graph.Base('G'), aligned_to=nid(0)),
            graph.Node(node_id=nid(2), base=graph.Base('C'), aligned_to=nid(3)),
            graph.Node(node_id=nid(3), base=graph.Base('G'), aligned_to=nid(2)),
            graph.Node(node_id=nid(4), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(5), base=graph.Base('A'), aligned_to=nid(6)),
            graph.Node(node_id=nid(6), base=graph.Base('C'), aligned_to=nid(5)),

            graph.Node(node_id=nid(7), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(8), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(9), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(10), base=graph.Base('T'), aligned_to=None),

            graph.Node(node_id=nid(11), base=graph.Base(self.missing_n.value), aligned_to=None),
            graph.Node(node_id=nid(12), base=graph.Base(self.missing_n.value), aligned_to=None),
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 2, 4, 5, 8, 9, 10])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [1, 3, 4, 6, 7, 11, 12])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            missings.ConstBaseProvider(self.missing_n),
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_3_missing_two_sequences_middle(self):
        maf_path = self.maf_files_dir.joinpath("test_3_missing_two_sequences_middle.maf")

        expected_nodes = [
            # block 0
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=nid(1)),
            graph.Node(node_id=nid(1), base=graph.Base('G'), aligned_to=nid(0)),
            graph.Node(node_id=nid(2), base=graph.Base('C'), aligned_to=None),

            # missing seq1
            graph.Node(node_id=nid(3), base=graph.Base(self.missing_n.value), aligned_to=None),
            graph.Node(node_id=nid(4), base=graph.Base(self.missing_n.value), aligned_to=None),

            # missing seq2
            graph.Node(node_id=nid(5), base=graph.Base(self.missing_n.value), aligned_to=None),
            graph.Node(node_id=nid(6), base=graph.Base(self.missing_n.value), aligned_to=None),

            # block 1
            graph.Node(node_id=nid(7), base=graph.Base('C'), aligned_to=nid(8)),
            graph.Node(node_id=nid(8), base=graph.Base('G'), aligned_to=nid(7)),
            graph.Node(node_id=nid(9), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(10), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(11), base=graph.Base('T'), aligned_to=None)
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 2, 3, 4, 8, 10, 11])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [1, 5, 6, 7, 9, 10, 11])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            missings.ConstBaseProvider(self.missing_n),
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_4_missing_one_sequence_middle(self):
        maf_path = self.maf_files_dir.joinpath("test_4_missing_one_sequence_middle.maf")

        expected_nodes = [
            # block 0
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=nid(1)),
            graph.Node(node_id=nid(1), base=graph.Base('G'), aligned_to=nid(0)),
            graph.Node(node_id=nid(2), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(3), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(4), base=graph.Base('A'), aligned_to=nid(5)),
            graph.Node(node_id=nid(5), base=graph.Base('G'), aligned_to=nid(4)),

            # missing se2
            graph.Node(node_id=nid(6), base=graph.Base(self.missing_n.value), aligned_to=None),
            graph.Node(node_id=nid(7), base=graph.Base(self.missing_n.value), aligned_to=None),

            graph.Node(node_id=nid(8), base=graph.Base('A'), aligned_to=nid(9)),
            graph.Node(node_id=nid(9), base=graph.Base('G'), aligned_to=nid(8)),
            graph.Node(node_id=nid(10), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(11), base=graph.Base('T'), aligned_to=None),

        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 2, 3, 4, 9, 10, 11])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [1, 5, 6, 7, 8, 10, 11])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            missings.ConstBaseProvider(self.missing_n),
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_5_missing_one_reverted_sequence_middle_1_1(self):
        maf_path = self.maf_files_dir.joinpath("test_5_missing_one_reverted_sequence_middle_1_1.maf")

        expected_nodes = [
            # block 0
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=nid(1)),
            graph.Node(node_id=nid(1), base=graph.Base('G'), aligned_to=nid(0)),
            graph.Node(node_id=nid(2), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(3), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(4), base=graph.Base('A'), aligned_to=nid(5)),
            graph.Node(node_id=nid(5), base=graph.Base('G'), aligned_to=nid(4)),

            # missing seq2, on edge (1,1)
            graph.Node(node_id=nid(6), base=graph.Base(self.missing_n.value), aligned_to=None),
            graph.Node(node_id=nid(7), base=graph.Base(self.missing_n.value), aligned_to=None),

            # block 1
            graph.Node(node_id=nid(8), base=graph.Base('A'), aligned_to=nid(9)),
            graph.Node(node_id=nid(9), base=graph.Base('G'), aligned_to=nid(8)),
            graph.Node(node_id=nid(10), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(11), base=graph.Base('T'), aligned_to=None)
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 2, 3, 4, 9, 10, 11])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [1, 5, 6, 7])]),
                                graph.SeqPath([*map(nid, [8, 10, 11])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            missings.ConstBaseProvider(self.missing_n),
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_6_missing_one_reverted_sequence_middle_minus1_1(self):
        maf_path = self.maf_files_dir.joinpath("test_6_missing_one_reverted_sequence_middle_minus1_1.maf")

        expected_nodes = [
            # block 1 because it is first in DAG and reverted
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(1), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(2), base=graph.Base('C'), aligned_to=nid(3)),
            graph.Node(node_id=nid(3), base=graph.Base('T'), aligned_to=nid(2)),

            # missing seq2, on edge (-1,1)
            graph.Node(node_id=nid(4), base=graph.Base(self.missing_n.value), aligned_to=None),
            graph.Node(node_id=nid(5), base=graph.Base(self.missing_n.value), aligned_to=None),

            graph.Node(node_id=nid(6), base=graph.Base('A'), aligned_to=nid(7)),
            graph.Node(node_id=nid(7), base=graph.Base('C'), aligned_to=nid(6)),
            graph.Node(node_id=nid(8), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(9), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(10), base=graph.Base('A'), aligned_to=nid(11)),
            graph.Node(node_id=nid(11), base=graph.Base('C'), aligned_to=nid(10)),

        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 1, 2])]),
                                graph.SeqPath([*map(nid, [6, 8, 9, 10])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 1, 3, 4, 5, 7, 11])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            missings.ConstBaseProvider(self.missing_n),
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_7_missing_one_reverted_sequence_middle_minus1_minus1(self):
        maf_path = self.maf_files_dir.joinpath("test_7_missing_one_reverted_sequence_middle_minus1_minus1.maf")

        expected_nodes = [
            # block 0
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(1), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(2), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(3), base=graph.Base('A'), aligned_to=None),

            # missing seq2
            graph.Node(node_id=nid(4), base=graph.Base(self.missing_n.value), aligned_to=None),
            graph.Node(node_id=nid(5), base=graph.Base(self.missing_n.value), aligned_to=None),

            # block 1
            graph.Node(node_id=nid(6), base=graph.Base('A'), aligned_to=nid(7)),
            graph.Node(node_id=nid(7), base=graph.Base('G'), aligned_to=nid(6)),
            graph.Node(node_id=nid(8), base=graph.Base('C'), aligned_to=nid(9)),
            graph.Node(node_id=nid(9), base=graph.Base('G'), aligned_to=nid(8)),
            graph.Node(node_id=nid(10), base=graph.Base('C'), aligned_to=nid(11)),
            graph.Node(node_id=nid(11), base=graph.Base('T'), aligned_to=nid(10)),
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 3, 7, 9, 11])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 1, 4, 5, 6, 8, 10])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            missings.ConstBaseProvider(self.missing_n),
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)


if __name__ == '__main__':
    unittest.main()



import unittest
from pathlib import Path

from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.pangenome import builder
from pangtreebuild.pangenome.parameters import missings
from pangtreebuild.tools import pathtools


def nid(x): return graph.NodeID(x)


def bid(x): return graph.BlockID(x)


class DAGMaf2PoagraphFakeFastaProviderTests(unittest.TestCase):

    def setUp(self):
        metadata_path = Path(__file__).parent.joinpath("../seq_metadata.csv").resolve()
        self.metadatacsv = msa.MetadataCSV(pathtools.get_file_content_stringio(metadata_path), metadata_path)
        self.maf_files_dir = Path(__file__).parent.joinpath("maf_files_with_cycles_or_reversion").resolve()
        self.fasta_provider = missings.ConstBaseProvider(missings.MissingBase())

    def test_00_simple(self):
        maf_path = self.maf_files_dir.joinpath("test_0_simple.maf")
        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(1), base=graph.Base('C'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(2), base=graph.Base('T'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(3), base=graph.Base('G'), aligned_to=None, block_id=bid(0)),

            graph.Node(node_id=nid(4), base=graph.Base('A'), aligned_to=None, block_id=bid(1)),
            graph.Node(node_id=nid(5), base=graph.Base('C'), aligned_to=None, block_id=bid(1)),
            graph.Node(node_id=nid(6), base=graph.Base('T'), aligned_to=None, block_id=bid(1)),
            graph.Node(node_id=nid(7), base=graph.Base('G'), aligned_to=None, block_id=bid(1)),
            graph.Node(node_id=nid(8), base=graph.Base('A'), aligned_to=None, block_id=bid(1)),
            graph.Node(node_id=nid(9), base=graph.Base('A'), aligned_to=None, block_id=bid(1))
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 3, 4, 5, 6, 7, 8])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'})),
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_01_reversed_seq_in_one_block(self):
        maf_path = self.maf_files_dir.joinpath("test_1_reversed_seq_in_one_block.maf")
        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=nid(1), block_id=bid(0)),
            graph.Node(node_id=nid(1), base=graph.Base('G'), aligned_to=nid(0), block_id=bid(0)),
            graph.Node(node_id=nid(2), base=graph.Base('C'), aligned_to=nid(3), block_id=bid(0)),
            graph.Node(node_id=nid(3), base=graph.Base('T'), aligned_to=nid(2), block_id=bid(0)),

            graph.Node(node_id=nid(4), base=graph.Base('G'), aligned_to=nid(5), block_id=bid(1)),
            graph.Node(node_id=nid(5), base=graph.Base('T'), aligned_to=nid(4), block_id=bid(1)),
            graph.Node(node_id=nid(6), base=graph.Base('A'), aligned_to=nid(7), block_id=bid(1)),
            graph.Node(node_id=nid(7), base=graph.Base('C'), aligned_to=nid(6), block_id=bid(1)),
            graph.Node(node_id=nid(8), base=graph.Base('C'), aligned_to=None, block_id=bid(1)),
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0, 2, 4, 7, 8])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [1, 3])]),
                                graph.SeqPath([*map(nid, [5, 6])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'})),
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_02_seq_starts_in_second_block(self):
        maf_path = self.maf_files_dir.joinpath(
                        "test_2_seq_starts_in_second_block.maf")

        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('C'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(1), base=graph.Base('T'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(2), base=graph.Base('G'), aligned_to=None, block_id=bid(0)),

            graph.Node(node_id=nid(3), base=graph.Base('T'), aligned_to=None, block_id=bid(1)),

            graph.Node(node_id=nid(4), base=graph.Base('G'), aligned_to=nid(5), block_id=bid(2)),
            graph.Node(node_id=nid(5), base=graph.Base('T'), aligned_to=nid(4), block_id=bid(2)),
            graph.Node(node_id=nid(6), base=graph.Base('A'), aligned_to=None, block_id=bid(2)),
            graph.Node(node_id=nid(7), base=graph.Base('A'), aligned_to=None, block_id=bid(2)),
            graph.Node(node_id=nid(8), base=graph.Base('C'), aligned_to=None, block_id=bid(2)),

        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [1, 2, 3])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 5, 7])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [3, 4, 6, 8])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_03_edge_not_from_last_node_in_block(self):
        maf_path = self.maf_files_dir.joinpath("test_3_edge_not_from_last_node_in_block.maf")

        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(1), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(2), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(3), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(4), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(5), base=graph.Base('A'), aligned_to=nid(6)),
            graph.Node(node_id=nid(6), base=graph.Base('T'), aligned_to=nid(5)),
            graph.Node(node_id=nid(7), base=graph.Base('C'), aligned_to=None),
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 1, 3, 4, 6, 7])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 3, 5])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_04_single_block_no_nucleotides(self):
        maf_path = self.maf_files_dir.joinpath(
                        "test_4_single_block_no_nucleotides.maf")

        expected_nodes = []

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_05_single_block_single_nucletodide(self):
        maf_path = self.maf_files_dir.joinpath(
                        "test_5_single_block_single_nucletodide.maf")

        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=None, block_id=bid(0))
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [graph.SeqPath([*map(nid, [0])])],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_06_1st_block_separates_into_2_branches_which_connect_in_3rd_block(self):
        maf_path = self.maf_files_dir.joinpath(
                        "test_6_1st_block_separates_into_2_branches_which_connect_in_3rd_block.maf")

        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=nid(1)),
            graph.Node(node_id=nid(1), base=graph.Base('C'), aligned_to=nid(2)),
            graph.Node(node_id=nid(2), base=graph.Base('G'), aligned_to=nid(0)),
            graph.Node(node_id=nid(3), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(4), base=graph.Base('A'), aligned_to=nid(5)),
            graph.Node(node_id=nid(5), base=graph.Base('T'), aligned_to=nid(4)),

            graph.Node(node_id=nid(6), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(7), base=graph.Base('G'), aligned_to=None),

            graph.Node(node_id=nid(8), base=graph.Base('C'), aligned_to=nid(9)),
            graph.Node(node_id=nid(9), base=graph.Base('G'), aligned_to=nid(10)),
            graph.Node(node_id=nid(10), base=graph.Base('T'), aligned_to=nid(8)),
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0, 3, 4, 8])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [1, 3, 5, 6, 7, 9])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [2, 3, 5, 10])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_07_inactive_edges_due_to_reversed_seqs(self):
        maf_path = self.maf_files_dir.joinpath("test_7_inactive_edges_due_to_reversed_seqs.maf")

        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(1), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(2), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(3), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(4), base=graph.Base('A'), aligned_to=nid(5)),
            graph.Node(node_id=nid(5), base=graph.Base('G'), aligned_to=nid(4)),
            graph.Node(node_id=nid(6), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(7), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(8), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(9), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(10), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(11), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(12), base=graph.Base('A'), aligned_to=nid(13)),
            graph.Node(node_id=nid(13), base=graph.Base('C'), aligned_to=nid(12)),
            graph.Node(node_id=nid(14), base=graph.Base('A'), aligned_to=nid(15)),
            graph.Node(node_id=nid(15), base=graph.Base('T'), aligned_to=nid(14)),
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 4, 6, 8, 9, 10, 11, 12, 14])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 1, 2])]),
                                graph.SeqPath([*map(nid, [3, 5])]),
                                graph.SeqPath([*map(nid, [6, 7, 8, 9, 10, 11, 13, 15])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [graph.SeqPath([*map(nid, [0, 2])]),
                                graph.SeqPath([*map(nid, [3, 5])]),
                                graph.SeqPath([*map(nid, [6, 7, 8, 9, 11, 12, 15])])],
                               graph.SequenceMetadata({'group': '2'})),
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_08_reversed_block(self):
        maf_path = self.maf_files_dir.joinpath("test_8_reversed_block.maf")

        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(1), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(2), base=graph.Base('T'), aligned_to=None),
            # next block is reversed because it was converted to dag
            graph.Node(node_id=nid(3), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(4), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(5), base=graph.Base('A'), aligned_to=nid(6)),
            graph.Node(node_id=nid(6), base=graph.Base('G'), aligned_to=nid(5)),
            graph.Node(node_id=nid(7), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(8), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(9), base=graph.Base('T'), aligned_to=None),
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 1, 3, 4, 5, 7, 8, 9])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 3, 4, 6, 7, 8, 9])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 3, 4, 6, 7, 9])])],
                               graph.SequenceMetadata({'group': '2'})),
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_09_inactive_edges_but_all_strands_plus(self):
        maf_path = self.maf_files_dir.joinpath("test_9_inactive_edges_but_all_strands_plus.maf")

        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(1), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(2), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(3), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(4), base=graph.Base('G'), aligned_to=None),

            graph.Node(node_id=nid(5), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(6), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(7), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(8), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(9), base=graph.Base('G'), aligned_to=None),

            graph.Node(node_id=nid(10), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(11), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(12), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(13), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(14), base=graph.Base('G'), aligned_to=None),

            graph.Node(node_id=nid(15), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(16), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(17), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(18), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(19), base=graph.Base('G'), aligned_to=None),
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 3, 4, 10, 11, 12, 13, 14])]),
                                graph.SeqPath([*map(nid, [5, 6, 7, 8, 9, 15, 16, 17, 18, 19])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,
                                                          13, 14, 15, 16, 17, 18, 19])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'})),
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_10_parallel_blocks_1st_and_2nd_merge_into_3rd(self):
        maf_path = self.maf_files_dir.joinpath("test_10_parallel_blocks_1st_and_2nd_merge_into_3rd.maf")

        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('G'), aligned_to=nid(1)),
            graph.Node(node_id=nid(1), base=graph.Base('T'), aligned_to=nid(0)),
            graph.Node(node_id=nid(2), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(3), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(4), base=graph.Base('C'), aligned_to=nid(5)),
            graph.Node(node_id=nid(5), base=graph.Base('G'), aligned_to=nid(4)),
            graph.Node(node_id=nid(6), base=graph.Base('C'), aligned_to=None),

            graph.Node(node_id=nid(7), base=graph.Base('A'), aligned_to=None),
            graph.Node(node_id=nid(8), base=graph.Base('C'), aligned_to=None),
            graph.Node(node_id=nid(9), base=graph.Base('T'), aligned_to=None),
            graph.Node(node_id=nid(10), base=graph.Base('G'), aligned_to=None),
            graph.Node(node_id=nid(11), base=graph.Base('G'), aligned_to=None),

            graph.Node(node_id=nid(12), base=graph.Base('C'), aligned_to=nid(13)),
            graph.Node(node_id=nid(13), base=graph.Base('G'), aligned_to=nid(12)),
            graph.Node(node_id=nid(14), base=graph.Base('C'), aligned_to=nid(15)),
            graph.Node(node_id=nid(15), base=graph.Base('G'), aligned_to=nid(16)),
            graph.Node(node_id=nid(16), base=graph.Base('T'), aligned_to=nid(14)),
            graph.Node(node_id=nid(17), base=graph.Base('A'), aligned_to=nid(18)),
            graph.Node(node_id=nid(18), base=graph.Base('T'), aligned_to=nid(17)),
            graph.Node(node_id=nid(19), base=graph.Base('A'), aligned_to=nid(20)),
            graph.Node(node_id=nid(20), base=graph.Base('C'), aligned_to=nid(19)),
            graph.Node(node_id=nid(21), base=graph.Base('C'), aligned_to=nid(22)),
            graph.Node(node_id=nid(22), base=graph.Base('G'), aligned_to=nid(21)),
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [7, 8, 9, 10, 11, 12, 15, 18, 19, 21])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [7, 8, 9, 10, 11, 12, 15, 18, 19, 21])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 2, 3, 4, 6, 13, 16, 17, 20, 21])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [graph.SeqPath([*map(nid, [1, 2, 3, 5, 6, 13, 14, 17, 20, 22])])],
                               graph.SequenceMetadata({'group': '2'})),
        }
        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        actual_poagraph, _ = builder.build_from_dagmaf(
            msa.Maf(pathtools.get_file_content_stringio(maf_path), maf_path),
            self.fasta_provider,
            self.metadatacsv)

        self.assertEqual(expected_poagraph, actual_poagraph)


if __name__ == '__main__':
    unittest.main()



import unittest
from pathlib import Path


from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import missings, msa
from pangtreebuild.tools import pathtools
from pangtreebuild.serialization import fasta
from pangtreebuild.affinity_tree import tree
from pangtreebuild.affinity_tree import parameters as at_params

def nid(x): return graph.NodeID(x)


def bid(x): return graph.Base(x)


class ToFASTATests(unittest.TestCase):

    def setUp(self):
        self.fasta_dir = Path(__file__).parent.joinpath('fasta_files/')

        poagraph_nodes = [graph.Node(node_id=nid(0), base=bid('A'), aligned_to=nid(1)),
                          graph.Node(node_id=nid(1), base=bid('G'), aligned_to=nid(0)),
                          graph.Node(node_id=nid(2), base=bid('C'), aligned_to=nid(3)),
                          graph.Node(node_id=nid(3), base=bid('G'), aligned_to=nid(2)),
                          graph.Node(node_id=nid(4), base=bid('A'), aligned_to=nid(5)),
                          graph.Node(node_id=nid(5), base=bid('T'), aligned_to=nid(4)),
                          graph.Node(node_id=nid(6), base=bid('G'), aligned_to=None),
                          graph.Node(node_id=nid(7), base=bid('G'), aligned_to=None),
                          graph.Node(node_id=nid(8), base=bid('A'), aligned_to=nid(9)),
                          graph.Node(node_id=nid(9), base=bid('C'), aligned_to=nid(10)),
                          graph.Node(node_id=nid(10), base=bid('G'), aligned_to=nid(11)),
                          graph.Node(node_id=nid(11), base=bid('T'), aligned_to=nid(8)),
                          graph.Node(node_id=nid(12), base=bid('A'), aligned_to=nid(13)),
                          graph.Node(node_id=nid(13), base=bid('C'), aligned_to=nid(12)),
                          graph.Node(node_id=nid(14), base=bid('T'), aligned_to=None),
                          graph.Node(node_id=nid(15), base=bid('A'), aligned_to=nid(16)),
                          graph.Node(node_id=nid(16), base=bid('C'), aligned_to=nid(17)),
                          graph.Node(node_id=nid(17), base=bid('G'), aligned_to=nid(15))]

        poagraph_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0, 2, 4, 6, 7, 8, 12, 14, 16])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [3, 4, 6, 7, 10, 12])]),
                                graph.SeqPath([*map(nid, [14, 17])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [graph.SeqPath([*map(nid, [11])]),
                                graph.SeqPath([*map(nid, [13, 14, 15])])],
                               graph.SequenceMetadata({'group': '1'})),
        }

        self.poagraph = graph.Poagraph(poagraph_nodes, poagraph_sequences)

    def test_1_sequences_fasta(self):
        expected_sequences_fasta_path = self.fasta_dir.joinpath("sequences.fasta")

        actual_sequences_fasta_content = fasta.poagraph_to_fasta(self.poagraph)
        expected_sequences_fasta_content = pathtools.get_file_content(expected_sequences_fasta_path)
        self.assertEqual(expected_sequences_fasta_content, actual_sequences_fasta_content)

    def test_2_consensuses_tree_fasta(self):
        expected_consensuses_fasta_path = self.fasta_dir.joinpath("consensuses.fasta")

        affinity_tree = tree.AffinityTree()
        affinity_tree.nodes = [
            # all members set
            tree.AffinityNode(id_=tree.AffinityNodeID(0),
                              parent=tree.AffinityNodeID(-1),
                              children=[tree.AffinityNodeID(1), tree.AffinityNodeID(2)],
                              sequences=[msa.SequenceID('seq0'),
                                         msa.SequenceID('seq1'),
                                         msa.SequenceID('seq2'),
                                         msa.SequenceID('seq3')],
                              mincomp=graph.Compatibility(0.5, at_params.P(1)),
                              compatibilities={msa.SequenceID('seq0'): graph.Compatibility(1.0, at_params.P(1)),
                                               msa.SequenceID('seq1'): graph.Compatibility(0.9, at_params.P(1)),
                                               msa.SequenceID('seq2'): graph.Compatibility(0.95, at_params.P(1)),
                                               msa.SequenceID('seq3'): graph.Compatibility(0.6, at_params.P(1))},
                              consensus=graph.SeqPath([nid(0), nid(2), nid(5), nid(6),
                                                       nid(10), nid(12), nid(13), nid(16)])),
            # no compatibilities to all, no mincomp
            tree.AffinityNode(id_=tree.AffinityNodeID(1),
                              parent=tree.AffinityNodeID(0),
                              sequences=[msa.SequenceID('seq0'),
                                         msa.SequenceID('seq1'),
                                         msa.SequenceID('seq2')],
                              consensus=graph.SeqPath([nid(0), nid(2), nid(3),
                                                       nid(6), nid(10), nid(11),
                                                       nid(13), nid(17)]))
        ]

        actual_consensuses_fasta_content = fasta.affinity_tree_to_fasta(self.poagraph, affinity_tree)
        expected_consensuses_fasta_content = pathtools.get_file_content(expected_consensuses_fasta_path)
        self.assertEqual(expected_consensuses_fasta_content, actual_consensuses_fasta_content)


if __name__ == '__main__':
    unittest.main()



"""Missing in MAF nucleotide or protein providers."""

import abc
from io import StringIO
import os
from pathlib import Path
import re
from typing import Dict, Optional
import zipfile

from Bio import Entrez
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.tools import pathtools, logprocess

detailed_logger = logprocess.get_logger("details")


class FastaProviderException(Exception):
    """Any exception connected with providing FASTA sequences."""

    pass


class FastaProvider(abc.ABC):
    """Base class for FASTA providers.

    MAF file usually contains not full sequences but only the aliggned parts.
    To build an exact poagraph full sequences must be retrieved
    from NCBI or FASTA file."""

    @abc.abstractmethod
    def get_base(self, seq_id: msa.SequenceID, i: int) -> graph.Base:
        """Returns base at position i in sequence identified by seq_id.

        Args:
            seq_id: Sequence ID of the sequence being currently complemented.
            i: position of the base in the sequence.

        Returns:
            The nucleotide or protein base.
        """

        pass

    @staticmethod
    def get_sequence_from_fasta(fasta_content: StringIO) -> str:
        """Returns sequence from FASTA file.

        FASTA file contains a header line and the sequence
        splitted into several lines. This method returns just the sequence.

        Args:
            fasta_content: Content of FASTA file.

        Returns:
            Return sequence saved in fasta structured file.
        """

        _ = fasta_content.readline()
        return fasta_content.read().replace('\n', '')


class MissingBase(object):
    """A constant symbol (default "?") replacing sequence missing symbols.

    Args:
        symbol: one-character long str to be saved as the missing symbol

    Attributes:
        value (str): the symbol as str
    """

    def __init__(self, symbol: Optional[str] = '?'):
        if len(symbol) != 1:
            raise ValueError('Missing symbol must be a single character.')
        self.value = symbol


class ConstBaseProvider(FastaProvider):
    """Missing base provider which returns always the same single character.

    Args:
        missing_base: the character to be always returned as missing base

    Attributes:
        missing_base (graph.Base): a constant sequence base
    """

    def __init__(self, missing_base: MissingBase):
        self.missing_base: graph.Base = graph.Base(missing_base.value)

    def get_base(self, seq_id: msa.SequenceID, i: int) -> graph.Base:
        """Returns const base symbol.

        Args:
            seq_id: ignored as the returned symbol is always the same.
            i: ignored as the returned symbol is always the same.

        Returns:
            The const symbol serving as missing symbol.
        """

        return self.missing_base


class FromFile(FastaProvider):
    """Missing base provider with FASTA file or zipped FASTA files source.

    Args:
        fasta: Path to the FASTA file or zipped FASTA files.
    """

    def __init__(self, fasta: Path):
        self._sequences: Dict[msa.SequenceID, str] = self._read_fastas(fasta)

    def get_base(self, seq_id: msa.SequenceID, i: int) -> graph.Base:
        """Returns base at position i in sequence identified by seq_id.

        Args:
            seq_id: ID of sequence present in given fasta file.
            i: position of the base to check in fasta file.

        Returns:
            The base present in seuqence seq_id at position i.
        """

        if seq_id not in self._sequences.keys():
            raise FastaProviderException(f"Wrong sequence id: {seq_id}. ")
        if i > len(self._sequences[seq_id]):
            raise FastaProviderException(f"""Index {i} is too large
                                          for sequence {seq_id}.""")
        return graph.Base(self._sequences[seq_id][i])

    def _read_fastas(self, fastas_file: Path) -> Dict[msa.SequenceID, str]:
        """Reads FASTA content to internal dictionary.

        Args:
            fastas_file: Path to the fasta file or zipped fasta files.

        Returns:
            Dictionary of sequence_id to sequence content.

        Raises:
            FastaProviderException: If there is no extension or the extension
            is unknown (other than zip/fasta/fna/faa.
        """

        fastas_file_suffix = fastas_file.suffix
        if fastas_file_suffix == '':
            raise FastaProviderException("""No file extension in fasta source.
                                        Append extension to file name.
                                        Supported: zip/fasta/fna/faa.""")
        else:
            fastas_file_extensions = fastas_file_suffix[1:]
            if fastas_file_extensions == "zip":
                return self._read_zip(fastas_file)
            elif fastas_file_extensions in ["fna", "fasta", "faa"]:
                return self._read_fasta(fastas_file)
            else:
                raise FastaProviderException("""Unknown fasta file format.
                                             Supported: zip/fasta/fna/faa.""")

    def _read_zip(self, zip_path: Path) -> Dict[msa.SequenceID, str]:
        """For given zip of fastas reads the sequences into a dictionary.

        Args:
            zip_path: Path tp the zip.

        Returns:
            Dictionary of sequence_id to sequence content.

        Raises:
            FastaProviderException: If no seuquences were found or
                                    the zipped fasta files were incorrect.

        """

        if not zipfile.is_zipfile(zip_path):
            raise FastaProviderException("Incorrect zip fasta source.")
        fastas_dict = {}
        with zipfile.ZipFile(zip_path) as myzip:
            zipinfolist = myzip.infolist()
            for zipinfo in zipinfolist:
                with myzip.open(zipinfo) as myfile:
                    fasta = myfile.read().decode("ASCII")
                    for record in SeqIO.parse(StringIO(fasta), "fasta"):
                        self._add_record_to_dict(record, fastas_dict)
        if len(fastas_dict) == 0:
            raise FastaProviderException("No sequences in zipped fastas or incorrect zipped files.")
        return fastas_dict

    def _read_fasta(self, fasta_path: Path) -> Dict[msa.SequenceID, str]:
        """Read sequences content from a single fasta file.

        Args:
            fasta_path: Path to the fasta file.

        Returns:
            Dictionary of sequence_id to sequence content.

        Raises:
            FastaProviderException: If no sequences were found or
                                    the zipped fasta files were incorrect.
        """
        fastas_dict = {}
        with open(fasta_path) as fasta_handle:
            for record in SeqIO.parse(fasta_handle, "fasta"):
                self._add_record_to_dict(record, fastas_dict)
        if len(fastas_dict) == 0:
            raise FastaProviderException("No sequences in zipped fastas or incorrect zipped files.")
        return fastas_dict

    def _add_record_to_dict(self,
                            record: SeqRecord,
                            fastas_dict: Dict[msa.SequenceID, str]) -> None:
        """Adds record the sequence to dict if the sequence is not empty.

        Args:
            record: The sequence to add.

        Raises:
            FastaProviderExecption: If record contains empty sequence or
                                    its ID is already present in the dict.
        """
        if len(record.seq) == 0:
            raise FastaProviderException("Empty sequence in FASTA. Provide the sequence or remove its header.")
        if msa.SequenceID(str(record.id)) in fastas_dict.keys():
            raise FastaProviderException("Incorrect fasta provided: sequences IDs are not unique.")
        fastas_dict[msa.SequenceID(str(record.id))] = record.seq


class FromNCBI(FastaProvider):
    """Missing base provider with NCBI as source.

    Args:
        use_cache:  Switch whether to save locally sequences downloaded
                    from NCBI and reuse them in further calls.
    """

    def __init__(self, use_cache: bool):
        Entrez.email = "paulinahyzy@gmail.com"
        self._fasta_disk_cache = _FastaDiskCache(Path(os.getcwd()))
        self._use_cache = use_cache
        self._sequences: Dict[msa.SequenceID, str] = {}

    def get_base(self, seq_id: msa.SequenceID, i: int) -> graph.Base:
        """Returns base at position i in sequence identified in NCBI
        by seq_id or sth similar.

        Args:
            seq_id: sequence_id, used as is or some guessing is performed
                    if no results available.
            i: position of the base to check in fasta file.

        Returns:
            The base present in seuqence sequence_id at position i.
        """
        if seq_id not in self._sequences.keys():
            sequence_is_cached = self._fasta_disk_cache.seq_is_cached(seq_id)
            if self._use_cache and sequence_is_cached:
                self._sequences[seq_id] = self._fasta_disk_cache.read(seq_id)
            elif self._use_cache and not sequence_is_cached:
                sequence = self._download_from_ncbi(seq_id)
                self._sequences[seq_id] = sequence
                self._fasta_disk_cache._save_to_cache(seq_id, sequence)
            else:
                self._sequences[seq_id] = self._download_from_ncbi(seq_id)

        return graph.Base(self._sequences[seq_id][i])

    def _download_from_ncbi(self, seq_id: msa.SequenceID) -> str:
        """Connects to NCBI and downloads full sequence identified by seq_id.

        Args:
            seq_id: Sequence ID in format [seq_id] or [seq_id]v[version].

        Returns:
            Sequence as str.

        Raises:
            ConnectionError: If any error while connecting NCBI appeared.
        """

        detailed_logger.info(f"Downloading from entrez sequence {seq_id}...")
        entrez_sequence_id = self._guess_ncbi_sequence_id(seq_id)
        try:
            handle = Entrez.efetch(db="nucleotide",
                                   id=entrez_sequence_id,
                                   rettype="fasta",
                                   retmode="text")
            fasta_content = FastaProvider.get_sequence_from_fasta(handle)
            return fasta_content
        except Exception as ex:
            raise ConnectionError(f"""Cannot download from Entrez
                                    sequence of ID: {seq_id}""") from ex

    def _guess_ncbi_sequence_id(self, seqid: msa.SequenceID) -> str:
        """Tries to guess sequence id by searching for version indicator.

        Args:
            seqid: Sequence ID in format [seq_id] or [seq_id]v[version].

        Returns:
            Guessed sequence ID.
        """

        detailed_logger.info(f"Guessing entrez sequence id...")
        version_indications = [*re.finditer('v[0-9]', seqid.value)]
        if len(version_indications) == 1:
            v_start = version_indications[0].span()[0]
            if v_start == len(seqid.value) - 2:
                guessed_entrez_name = ".".join([seqid.value[0:v_start],
                                                seqid.value[v_start+1:]])
            else:
                guessed_entrez_name = seqid.value
        else:
            guessed_entrez_name = seqid.value
        detailed_logger.info(f"{seqid} translated to {guessed_entrez_name}")
        return guessed_entrez_name


class _FastaDiskCache(object):
    """Manages caching sequences downloaded from NCBI.

    Args:
        parent_dir: Path to the directory where new directory
                    with cached sequences will be created.
    """

    def __init__(self, parent_dir: Path):
        self._parent_dir = parent_dir
        self._cache_dir = pathtools.get_child_path(parent_dir, ".fastacache")

    def seq_is_cached(self, seq_id: msa.SequenceID) -> bool:
        """Returns information whether sequence with seq_id is cached.

        Args:
            seq_id: Sequence ID.

        Returns:
            True if sequence is cached, False if not.
        """

        if not self._cache_dir_exists():
            return False
        expected_fasta_file_name = self._get_cached_filepath(seq_id)
        fasta_files_in_cache_dir = self._cache_dir.glob("*.fasta")
        if expected_fasta_file_name in [*fasta_files_in_cache_dir]:
            return True
        return False

    def _cache_dir_exists(self) -> bool:
        return pathtools.dir_exists(self._cache_dir)

    def _get_cached_filepath(self, seq_id: msa.SequenceID) -> Path:
        return pathtools.get_child_path(self._cache_dir, f"{seq_id}.fasta")

    def read(self, seq_id: msa.SequenceID) -> str:
        """Gets full sequence identified by seq_id from cached local file.

        Args:
            seq_id: Sequence ID.

        Returns:
            The sequence as str.
        """

        detailed_logger.info(f"Reading {seq_id} from cache...")
        cache_filepath = self._get_cached_filepath(seq_id)
        with open(cache_filepath) as fasta_handle:
            seq = SeqIO.read(fasta_handle, "fasta")
        return seq.seq

    def _create_cache_dir(self) -> None:
        if not self._cache_dir_exists():
            pathtools.create_dir(self._cache_dir)
            detailed_logger.info(".fastacache directory was created.")
        else:
            detailed_logger.warning(""".fastacache directory not created,
                                    as it already exists.""")

    def _save_to_cache(self, seq_id: msa.SequenceID, sequence: str) -> None:
        detailed_logger.info(f"Caching sequence {seq_id}...")
        if not self._cache_dir_exists():
            self._create_cache_dir()
        cache_filename = self._get_cached_filepath(seq_id)
        with open(cache_filename, 'w') as fasta_file_handle:
            SeqIO.write(SeqRecord(seq=Seq(sequence),
                                  id=seq_id.value,
                                  description="cached"),
                        fasta_file_handle,
                        "fasta")



import unittest
from pathlib import Path

from ddt import unpack, data, ddt
from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import missings, msa

@ddt
class FromNCBITests(unittest.TestCase):
    def setUp(self) -> None:
        self.fasta_dir = Path(__file__).parent.joinpath("fasta_ncbi/").resolve()
        self.fasta_provider = missings.FromNCBI(use_cache=False)

    # @unittest.skip("slow test - internet connection required")
    def test_0_get_10th_symbol_of_AB050936v1(self):
        sequence_id = msa.SequenceID("AB050936.1")
        actual_base = self.fasta_provider.get_base(sequence_id, 10)
        path = self.fasta_dir.joinpath("AB050936.1.fasta")
        expected_base = graph.Base(self.read_sequence(path)[10])
        self.assertEqual(expected_base, actual_base)

    # @unittest.skip("slow test - internet connection required")
    def test_1_download_AB050936v1(self):
        fasta_provider = missings.FromNCBI(use_cache=False)
        sequence_id = msa.SequenceID("AB050936.1")
        actual_sequence = fasta_provider._download_from_ncbi(sequence_id)
        p = self.fasta_dir.joinpath('AB050936.1.fasta')
        expected_sequence = self.read_sequence(p)
        self.assertEqual(expected_sequence, actual_sequence)

    # @unittest.skip("slow test - internet connection required")
    def test_2_failed_download(self):
        fasta_provider = missings.FromNCBI(use_cache=False)
        sequence_id = ""
        with self.assertRaises(Exception) as err:
            _ = fasta_provider._download_from_ncbi(msa.SequenceID(sequence_id))
            self.assertEqual(str(err), f"Cannot download from Entrez sequence of ID: {sequence_id}")

    @data((msa.SequenceID("plain"), "plain"),
          (msa.SequenceID("withv1"), "with.1"))
    @unpack
    def test_3_guess_entrez_id(self,
                               sequenceID: msa.SequenceID,
                               expected_guessed_entrez_id: str):
        fasta_provider = missings.FromNCBI(use_cache=False)
        actual_guessed_entrez_id = fasta_provider._guess_ncbi_sequence_id(sequenceID)

        self.assertEqual(expected_guessed_entrez_id, actual_guessed_entrez_id)

    @staticmethod
    def read_sequence(path: Path):
        with open(path) as fasta_file_hanlder:
            _ = fasta_file_hanlder.readline()
            return fasta_file_hanlder.read().upper().replace("\n", "")


if __name__ == '__main__':
    unittest.main()



from collections import namedtuple
from typing import List, Optional, Tuple, Dict

from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import msa


POSequenceInfo = namedtuple('POSequenceInfo', ['name',
                                               'nodes_count',
                                               'start_node',
                                               'weight',
                                               'consensus_id',
                                               'additional_info'])


def get_poagraph(po: msa.Po, metadataCsv: msa.MetadataCSV):
    """Returns poagraph elements based on multialignment PO file and metadata.

    Args:
        po: Multaalignment in PO format.
        metadataCsv: MetadataCSV.

    Returns:
        Tuple of poagraph elements.
    """

    po_lines = po.filecontent.readlines()
    sequences_info = _get_sequences_info_from_po(po_lines)
    initial_sequences = _init_sequences(sequences_info, metadataCsv)
    nodes, sequences = _get_poagraph_paths_and_nodes(po_lines,
                                                     sequences_info,
                                                     initial_sequences)
    return nodes, sequences


def _extract_line_value(line: str) -> str:
    splitted_line = line.split('=')
    if len(splitted_line) < 2:
        raise Exception("Incorrect line format. Nothing after \"=\" found.")
    return splitted_line[1].strip()


def _init_sequences(sequences_info: Dict[int, POSequenceInfo],
                    metadata: Optional[msa.MetadataCSV]) -> \
        Dict[msa.SequenceID, graph.Sequence]:
    metadata_sequences_ids = metadata.get_all_sequences_ids() \
                              if metadata else []
    po_sequences_ids = [seq_info.name for seq_info in sequences_info.values()]
    initial_sequences = {seq_id: graph.Sequence(seqid=seq_id,
                                                paths=[],
                                                seqmetadata=metadata.get_sequence_metadata(seq_id)
                                                if metadata else {})
                         for seq_id in set(po_sequences_ids + metadata_sequences_ids)}

    return initial_sequences


def _get_sequences_info_from_po(po_lines: List[str]) -> \
        Dict[int, POSequenceInfo]:
    paths_count = int(_extract_line_value(po_lines[4]))
    sequences_info = {}
    po_seq_id = 0
    for i in range(5, 5 + paths_count * 2, 2):
        path_name = _extract_line_value(po_lines[i])
        detailed_info_line = po_lines[i+1]
        detailed_info = _extract_line_value(detailed_info_line).split(' ')
        if len(detailed_info) < 5:
            raise Exception(f"""Expeceted SOURCEINFO=[5 parameters].
                                Got {detailed_info_line} instead.""")
        sequences_info[po_seq_id] = POSequenceInfo(name=msa.SequenceID(path_name),
                                                   nodes_count=detailed_info[0],
                                                   start_node=detailed_info[1],
                                                   weight=detailed_info[2],
                                                   consensus_id=detailed_info[3],
                                                   additional_info="".join(detailed_info[4:]))
        po_seq_id += 1
    return sequences_info


def _get_poagraph_paths_and_nodes(po_lines: List[str],
                                  sequences_info: Dict[int, POSequenceInfo],
                                  sequences: Dict[msa.SequenceID, graph.Sequence]) -> \
        Tuple[List[graph.Node], Dict[msa.SequenceID, graph.Sequence]]:
    nodes_count = int(_extract_line_value(po_lines[3]))
    paths_count = int(_extract_line_value(po_lines[4]))
    nodes: List[graph.Node] = [None] * nodes_count
    node_id = 0
    for i in range(5 + paths_count * 2, 5 + paths_count * 2 + nodes_count):
        node_line = po_lines[i]
        base = graph.Base(node_line[0].upper())
        in_nodes, po_sequences_ids, aligned_to = _extract_node_parameters(node_line)
        sequences_ids = [sequences_info[po_sequences_id].name
                         for po_sequences_id in po_sequences_ids]
        nodes[node_id] = graph.Node(graph.NodeID(node_id),
                                    base,
                                    graph.NodeID(aligned_to))
        for seq_id in sequences_ids:
            if len(sequences[seq_id].paths) == 1:
                sequences[seq_id].paths[0].append(graph.NodeID(node_id))
            else:
                sequences[seq_id].paths.append(graph.SeqPath([graph.NodeID(node_id)]))
        node_id += 1
    return nodes, sequences


def _extract_node_parameters(line: str) -> Tuple[List[int], List[int], int]:
    line = line[2:]
    line_iter = iter(line)

    node_parameters = {'L': [], 'S': [], 'A': [], '': []}
    current_node_parameter = next(line_iter)
    number_start = 1
    number_end = 1
    for i, c in enumerate(line_iter):
        if c == 'L' or c == 'S' or c == 'A':
            node_parameters[current_node_parameter].append(int(line[number_start: number_end]))
            current_node_parameter = c
            number_start = i + 2
            number_end = i + 2
        else:
            number_end += 1
    node_parameters[current_node_parameter].append(int(line[number_start: number_end]))
    aligned_to = node_parameters['A'][0] if node_parameters['A'] else None
    return node_parameters['L'], node_parameters['S'], aligned_to



import unittest
from pathlib import Path

from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import missings, msa


class FromFileFastaProviderFastaTests(unittest.TestCase):
    def setUp(self) -> None:
        self.fasta_dir = Path(__file__).parent.joinpath("files_fasta/").resolve()

    @staticmethod
    def read_sequence(path: Path):
        with open(path) as fasta_file_hanlder:
            _ = fasta_file_hanlder.readline()
            return fasta_file_hanlder.read().upper().replace("\n", "")

    def raise_error_if_unequal(self,
                               sequence_id: msa.SequenceID,
                               expected_sequence: str,
                               fasta_provider: missings.FromFile) -> None:
        for i, expected_symbol in enumerate(expected_sequence):
            expected_base = graph.Base(expected_symbol)
            actual_base = fasta_provider.get_base(sequence_id, i)
            self.assertEqual(expected_base, actual_base)

    def test_1_one_sequence(self):
        fasta_path = self.fasta_dir.joinpath("test_1_one_sequence.fasta")
        fasta_provider = missings.FromFile(Path(fasta_path))

        sequence_id = msa.SequenceID("seq1")
        expected_sequence = self.read_sequence(fasta_path)

        self.raise_error_if_unequal(sequence_id,
                                    expected_sequence,
                                    fasta_provider)

    def test_2_three_sequences(self):
        fasta_path = self.fasta_dir.joinpath("test_2_three_sequences.fasta")

        fasta_provider = missings.FromFile(Path(fasta_path))

        sequence_id_1 = msa.SequenceID("seq1")
        self.raise_error_if_unequal(sequence_id_1,
                                    "ACTGGGTGGGA",
                                    fasta_provider)

        sequence_id_2 = msa.SequenceID("seq2")
        self.raise_error_if_unequal(sequence_id_2,
                                    "AA",
                                    fasta_provider)

        sequence_id_3 = msa.SequenceID("seq3")
        self.raise_error_if_unequal(sequence_id_3,
                                    "GT",
                                    fasta_provider)

    def test_3_empty_sequence_name(self):
        fasta_path = self.fasta_dir.joinpath("test_3_empty_sequence_name.fasta")

        with self.assertRaises(Exception) as exp:
            _ = missings.FromFile(Path(fasta_path))

        expected_message = ("No sequences in zipped fastas or incorrect zipped files.")
        actual_message = str(exp.exception)
        self.assertEqual(expected_message, actual_message)

    def test_4_empty_sequence(self):
        fasta_path = self.fasta_dir.joinpath("test_4_empty_sequence.fasta")

        with self.assertRaises(Exception) as exp:
            _ = missings.FromFile(Path(fasta_path))

        expected_message = "Empty sequence in FASTA. Provide the sequence or remove its header."
        actual_message = str(exp.exception)
        self.assertEqual(expected_message, actual_message)

    def test_5_empty_fasta(self):
        fasta_path = self.fasta_dir.joinpath("test_5_empty_fasta.fasta")

        with self.assertRaises(Exception) as exp:
            _ = missings.FromFile(Path(fasta_path))

        expected_message = "No sequences in zipped fastas or incorrect zipped files."
        actual_message = str(exp.exception)
        self.assertEqual(expected_message, actual_message)

    def test_6_repeated_sequences(self):
        fasta_path = self.fasta_dir.joinpath("test_6_repeated_sequences.fasta")

        with self.assertRaises(Exception) as exp:
            _ = missings.FromFile(Path(fasta_path))

        expected_message = "Incorrect fasta provided: sequences IDs are not unique."
        actual_message = str(exp.exception)
        self.assertEqual(expected_message, actual_message)


if __name__ == '__main__':
    unittest.main()



from typing import List


class DAGMafNode:
    """Single node of Multialignment DAG."""

    def __init__(self, block_id, alignment, orient, order, out_edges):
        self.id = block_id
        self.alignment = alignment
        self.orient = orient  # orientation relative to parent
        self.order = order
        self.out_edges = out_edges


class DAGMaf:
    """MAF multialignment converted to DAG."""

    def __init__(self, dagmaf_nodes: List[DAGMafNode]):
        self.dagmaf_nodes = dagmaf_nodes



from collections import namedtuple
from typing import Tuple, List, NewType, Optional, Dict

from pangtreebuild.mafgraph.graph import Block
from pangtreebuild.mafgraph.graph.Arc import Arc
from pangtreebuild.mafgraph.mafreader import start_position
from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome import DAGMaf
from pangtreebuild.pangenome.parameters import missings
from pangtreebuild.pangenome.parameters import msa

from pangtreebuild.tools import logprocess


global_logger = logprocess.get_global_logger()
detailed_logger = logprocess.get_logger("details")


class PoagraphBuildException(Exception):
    """Any exception connected with building poagraph."""

    pass


MafSequenceID = NewType('MafSequenceID', str)

SequenceInfo = namedtuple('SequenceInfo', ['block_id',
                                           'start',
                                           'strand',
                                           'size',
                                           'srcSize',
                                           'orient'])

Edge = namedtuple('Edge', ['seq_id',
                           'from_block_id',
                           'to_block_id',
                           'last_node_id'])


class _BuildState:
    def __init__(self,
                 initial_nodes: List[graph.Node],
                 initial_sequences: Dict[msa.SequenceID, graph.Sequence],
                 initial_edges: Dict[msa.SequenceID, List[Edge]],
                 seqs_info: Dict[msa.SequenceID, List[SequenceInfo]],
                 initial_column_id: graph.ColumnID,
                 fasta_provider: missings.FastaProvider):
        self.nodes: List[graph.Node] = initial_nodes
        self.sequences: Dict[msa.SequenceID, graph.Sequence] = initial_sequences
        self.free_edges: Dict[msa.SequenceID, List[Edge]] = initial_edges
        self.seqs_info: Dict[msa.SequenceID, List[SequenceInfo]] = seqs_info
        self.column_id: graph.ColumnID = initial_column_id
        self.fasta_provider: missings.FastaProvider = fasta_provider


def get_poagraph(dagmaf: DAGMaf.DAGMaf,
                 fasta_provider: missings.FastaProvider,
                 metadata: Optional[msa.MetadataCSV]) -> \
        Tuple[List[graph.Node], Dict[msa.SequenceID, graph.Sequence]]:
    """Gets poagraph from given dagmaf using fasta_provider and metadata.

    Args:
        dagmaf: DagMaf that will be converted to Poagraph.
        fasta_provider: Provider of symbols missing in DagMaf.
        metadata: MetadataCSV.

    Returns:
        Tuple of poagraph elements.
    """

    sequences_in_dagmaf = _get_sequences_ids(dagmaf)
    build_state = _BuildState(initial_nodes=[],
                              initial_sequences=_init_sequences(sequences_in_dagmaf, metadata),
                              initial_edges=_init_free_edges(sequences_in_dagmaf),
                              seqs_info=_get_seqs_info(dagmaf, sequences_in_dagmaf),
                              initial_column_id=graph.ColumnID(-1),
                              fasta_provider=fasta_provider)

    _complement_starting_nodes(build_state)

    for i, mafnode in enumerate(dagmaf.dagmaf_nodes):
        _process_block(build_state, mafnode)

    return build_state.nodes, build_state.sequences


def _get_sequences_ids(dagmaf: DAGMaf.DAGMaf) -> List[msa.SequenceID]:
    return list({msa.SequenceID(seq.id)
                 for block in dagmaf.dagmaf_nodes
                 for seq in block.alignment})


def _init_sequences(sequences_in_dagmaf: List[msa.SequenceID],
                    metadata: Optional[msa.MetadataCSV]) -> \
        Dict[msa.SequenceID, graph.Sequence]:
    metadata_sequences_ids = metadata.get_all_sequences_ids() if metadata else []
    initial_sequences = {seq_id: graph.Sequence(seqid=seq_id,
                                                paths=[],
                                                seqmetadata=metadata.get_sequence_metadata(seq_id)
                                                if metadata else {})
                         for seq_id in set(sequences_in_dagmaf + metadata_sequences_ids)}

    return initial_sequences


def _init_free_edges(maf_sequences_ids: List[msa.SequenceID]) -> \
        Dict[msa.SequenceID, List[Edge]]:
    return {seq_id: [] for seq_id in maf_sequences_ids}


def _get_seqs_info(dagmaf: DAGMaf.DAGMaf,
                   sequences_in_dagmaf: List[msa.SequenceID]) -> \
        Dict[msa.SequenceID, List[SequenceInfo]]:
    seqs_info = {seq_id: [] for seq_id in sequences_in_dagmaf}

    for n in dagmaf.dagmaf_nodes:
        for seq in n.alignment:
            seqs_info[msa.SequenceID(seq.id)].append(SequenceInfo(block_id=graph.BlockID(n.id),
                                                                  start=start_position(seq),
                                                                  strand=seq.annotations["strand"],
                                                                  size=seq.annotations["size"],
                                                                  srcSize=seq.annotations["srcSize"],
                                                                  orient=n.orient))
    absents_sequences: List[msa.SequenceID] = []
    for seq_id, seq_info_list in seqs_info.items():
        if seq_info_list:
            seqs_info[seq_id] = sorted(seq_info_list, key=lambda si: si.start)
        else:
            absents_sequences.append(seq_id)
    for seq_id in absents_sequences:
        del seqs_info[seq_id]
    return seqs_info


def _complement_starting_nodes(build_state: _BuildState) -> None:
    for seq_id, seq_info_list in build_state.seqs_info.items():
        first_block_sinfo = seq_info_list[0]
        if first_block_sinfo.start != 0:
            _complement_sequence_starting_nodes(build_state,
                                                seq_id,
                                                first_block_sinfo)


def _complement_sequence_starting_nodes(build_state: _BuildState,
                                        seq_id: msa.SequenceID,
                                        first_block_sinfo: SequenceInfo) -> \
        None:
    current_node_id: graph.NodeID = _get_max_node_id(build_state.nodes)
    column_id = -first_block_sinfo.start
    join_with = None
    for i in range(first_block_sinfo.start):
        current_node_id += 1
        missing_nucleotide = _get_missing_nucleotide(build_state.fasta_provider, seq_id, i)
        build_state.nodes += [graph.Node(node_id=current_node_id,
                                         base=missing_nucleotide,
                                         column_id=column_id)]
        _add_node_to_sequence(build_state,
                              seq_id=seq_id,
                              join_with=join_with,
                              node_id=current_node_id)
        join_with = current_node_id
        column_id += 1
    build_state.free_edges[seq_id] += [Edge(seq_id=seq_id,
                                            from_block_id=None,
                                            to_block_id=first_block_sinfo.block_id,
                                            last_node_id=current_node_id)]


def _get_max_node_id(nodes: List[graph.Node]) -> graph.NodeID:
    return graph.NodeID(len(nodes) - 1)


def _get_missing_nucleotide(fasta_provider, seq_id: msa.SequenceID, i: int) -> graph.Base:
    return fasta_provider.get_base(seq_id, i)


def _add_node_to_sequence(build_state: _BuildState,
                          seq_id: msa.SequenceID,
                          join_with: graph.NodeID,
                          node_id: graph.NodeID) -> None:
    if len(build_state.sequences[seq_id].paths) == 0 or join_with is None:
        build_state.sequences[seq_id].paths.append(graph.SeqPath([node_id]))
    else:
        for path in build_state.sequences[seq_id].paths:
            if path[-1] == join_with:
                path.append(node_id)
                return

        raise PoagraphBuildException("No path with specified last node id.")


def _process_block(build_state: _BuildState, block: DAGMaf.DAGMafNode):
    current_node_id = _get_max_node_id(build_state.nodes)
    block_width = len(block.alignment[0].seq)
    paths_join_info = _get_paths_join_info(block, build_state.free_edges)

    build_state.column_id = _get_max_column_id(build_state.nodes)
    for col in range(block_width):
        build_state.column_id += 1
        sequence_name_to_nucleotide = {MafSequenceID(seq.id): seq[col]
                                       for seq in block.alignment}
        nodes_codes = _get_column_nucleotides_sorted_codes(sequence_name_to_nucleotide)
        column_nodes_ids = [current_node_id + i + 1 for i, _ in enumerate(nodes_codes)]
        for i, nucl in enumerate(nodes_codes):
            current_node_id += 1
            maf_seqs_id = [seq_id for seq_id, n in sequence_name_to_nucleotide.items() if n == nucl]
            build_state.nodes += [graph.Node(node_id=current_node_id,
                                             base=graph.Base(nucl),
                                             aligned_to=_get_next_aligned_node_id(i, column_nodes_ids),
                                             column_id=build_state.column_id,
                                             block_id=block.id)]

            for maf_seq_id in maf_seqs_id:
                seq_id = msa.SequenceID(maf_seq_id)
                _add_node_to_sequence(build_state, seq_id, paths_join_info[seq_id], current_node_id)
                paths_join_info[seq_id] = current_node_id

    _add_block_out_edges_to_free_edges(build_state, block, paths_join_info)
    _manage_endings(build_state, block, paths_join_info)


def _get_paths_join_info(block: Block,
                         free_edges: Dict[msa.SequenceID, List[Edge]]) -> \
        Dict[msa.SequenceID, Optional[graph.NodeID]]:
    paths_join_info: Dict[msa.SequenceID, Optional[graph.NodeID]] = dict()
    for seq in block.alignment:
        seq_id = msa.SequenceID(seq.id)
        paths_join_info[seq_id] = None
        for i, edge in enumerate(free_edges[seq_id]):
            if edge.to_block_id == block.id:
                paths_join_info[seq_id] = edge.last_node_id
    return paths_join_info


def _get_max_column_id(nodes: List[graph.Node]) -> graph.ColumnID:
    current_columns_ids = [node.column_id for node in nodes]
    return max(current_columns_ids) if current_columns_ids \
        else graph.ColumnID(-1)


def _get_column_nucleotides_sorted_codes(seq_to_nucl: Dict[msa.SequenceID, str]) -> \
        List[str]:
    return sorted(
        set(
            [nucleotide
             for nucleotide
             in seq_to_nucl.values()
             if nucleotide != '-']))


def _get_next_aligned_node_id(current_column_i, column_nodes_ids) -> \
        Optional[graph.NodeID]:
    if len(column_nodes_ids) > 1:
        return column_nodes_ids[(current_column_i + 1) % len(column_nodes_ids)]
    return None


def _add_block_out_edges_to_free_edges(build_state: _BuildState,
                                       block: Block,
                                       join_info: Dict[msa.SequenceID, graph.NodeID]):
    for edge in block.out_edges:
        _ = _get_correct_edge_type(edge)
        for seq in edge.sequences:
            seq_id = msa.SequenceID(seq[0].seq_id)
            last_node_id = _complement_sequence_middles_if_needed(build_state=build_state,
                                                                  block=block,
                                                                  edge=edge,
                                                                  seq=seq,
                                                                  last_node_id=join_info[seq_id])
            if last_node_id is not None:
                build_state.free_edges[seq_id].append(Edge(seq_id=seq_id,
                                                           from_block_id=block.id,
                                                           to_block_id=edge.to,
                                                           last_node_id=last_node_id))


def _get_correct_edge_type(edge: Arc) -> Tuple[int, int]:
    return edge.edge_type


def _complement_sequence_middles_if_needed(build_state: _BuildState,
                                           block: Block,
                                           edge: Arc,
                                           seq,
                                           last_node_id: graph.NodeID):
    seq_id = msa.SequenceID(seq[0].seq_id)
    left_block_sinfo, right_block_sinfo = _get_edge_sinfos(seqs_info=build_state.seqs_info,
                                                           from_block_id=block.id,
                                                           edge=edge,
                                                           seq_id=seq_id)
    if _complementation_not_needed(left_block_sinfo, right_block_sinfo):
        if edge.edge_type == (1, -1):
            return last_node_id
        else:
            return None
    else:
        current_node_id = _get_max_node_id(build_state.nodes)
        column_id = build_state.column_id
        if left_block_sinfo.start < right_block_sinfo.start:
            last_pos = left_block_sinfo.start + left_block_sinfo.size - 1
            next_pos = right_block_sinfo.start
        else:
            last_pos = right_block_sinfo.start + right_block_sinfo.size - 1
            next_pos = left_block_sinfo.start

        join_with = last_node_id if _should_join_with_last_node(edge.edge_type) else None
        for i in range(last_pos + 1, next_pos):
            column_id += 1
            current_node_id += 1
            missing_nucleotide = _get_missing_nucleotide(build_state.fasta_provider, seq_id, i)
            build_state.nodes += [graph.Node(node_id=current_node_id,
                                             base=missing_nucleotide,
                                             aligned_to=None,
                                             column_id=column_id,
                                             block_id=None)]
            _add_node_to_sequence(build_state,
                                  seq_id=seq_id,
                                  join_with=join_with,
                                  node_id=current_node_id)
            join_with = current_node_id

        if _should_join_with_next_node(edge.edge_type):
            return current_node_id
        else:
            return None


def _get_edge_sinfos(seqs_info: Dict[msa.SequenceID, List[SequenceInfo]],
                     from_block_id: graph.BlockID,
                     edge: Arc,
                     seq_id: msa.SequenceID) -> \
        Tuple[SequenceInfo, SequenceInfo]:
    left_seq_info, right_seq_info = None, None
    for sinfo in seqs_info[seq_id]:
        if sinfo.block_id == from_block_id:
            left_seq_info = sinfo
        if sinfo.block_id == edge.to:
            right_seq_info = sinfo
    if left_seq_info is None or right_seq_info is None:
        raise PoagraphBuildException(f"""SequenceInfos for edge cannot be None.
                                     Left block is {left_seq_info},
                                     right block is {right_seq_info}.""")
    return left_seq_info, right_seq_info


def _complementation_not_needed(left: SequenceInfo, right: SequenceInfo) -> \
        bool:
    return left.start + left.size == right.start or \
           right.start + right.size == left.start


def _should_join_with_last_node(edge_type: Tuple[int, int]) -> bool:
    if edge_type == (1, 1) or edge_type == (1, -1):
        return True
    elif edge_type == (-1, 1) or edge_type == (-1, -1):
        return False
    else:
        raise PoagraphBuildException("""Incorrect edge type.
                                     Cannot decide if sequence should be joined
                                     with complemented nucleotides.""")


def _should_join_with_next_node(edge_type: Tuple[int, int]) -> bool:
    if edge_type == (-1, 1) or edge_type == (1, -1) or edge_type == (-1, -1):
        return True
    elif edge_type == (1, 1):
        return False
    else:
        raise PoagraphBuildException("""Incorrect edge type. Cannot decide if
                                        complemented nucleotides must be joined
                                        with next block.""")


def _manage_endings(build_state: _BuildState,
                    block: Block,
                    join_info: Dict[msa.SequenceID, graph.NodeID]):
    sequences_ending_in_this_block = _get_ending_sequences(build_state.seqs_info, block)
    for seq_id in sequences_ending_in_this_block:
        block_sinfo: SequenceInfo = _get_sinfo(build_state.seqs_info[seq_id], block.id)
        if _sequence_not_complete(block_sinfo):
            last_node_id = _complement_sequence_middle_nodes(build_state,
                                                             seq_id=seq_id,
                                                             last_pos=block_sinfo.start + block_sinfo.size-1,
                                                             next_pos=block_sinfo.srcSize,
                                                             last_node_id=join_info[seq_id])
        else:
            last_node_id = join_info[seq_id]
        build_state.free_edges[seq_id].append(Edge(seq_id=seq_id,
                                                   from_block_id=block.id,
                                                   to_block_id=None,
                                                   last_node_id=last_node_id))


def _get_ending_sequences(seqs_info: Dict[msa.SequenceID, List[SequenceInfo]], block: Block) -> List[msa.SequenceID]:
    sequences_ending_in_this_block = []
    for seq_id, sinfo_list in seqs_info.items():
        last_block_sinfo = sinfo_list[-1]
        if last_block_sinfo.block_id == block.id:
            sequences_ending_in_this_block.append(seq_id)
    return sequences_ending_in_this_block


def _get_sinfo(seq_info: List[SequenceInfo], block_id: int) -> SequenceInfo:
    for sinfo in seq_info:
        if sinfo.block_id == block_id:
            return sinfo
    raise PoagraphBuildException(f"No sequences info for given block")


def _sequence_not_complete(last_block_sinfo: SequenceInfo) -> bool:
    if last_block_sinfo.strand == 1:
        return last_block_sinfo.start + last_block_sinfo.size != last_block_sinfo.srcSize
    elif last_block_sinfo.strand == -1:
        return last_block_sinfo.start != 0
    else:
        raise Exception("Unexpected strand value")


def _complement_sequence_middle_nodes(build_state: _BuildState,
                                      seq_id: msa.SequenceID,
                                      last_pos,
                                      next_pos,
                                      last_node_id: graph.NodeID) -> \
        graph.NodeID:
    current_node_id = _get_max_node_id(build_state.nodes)
    column_id = build_state.column_id
    join_with = last_node_id
    for i in range(last_pos+1, next_pos):
        column_id += 1
        current_node_id += 1
        missing_nucleotide = _get_missing_nucleotide(build_state.fasta_provider, seq_id, i)
        build_state.nodes += [graph.Node(node_id=current_node_id,
                                         base=missing_nucleotide,
                                         aligned_to=None,
                                         column_id=column_id,
                                         block_id=None)
                              ]
        _add_node_to_sequence(build_state,
                              seq_id=seq_id,
                              join_with=join_with,
                              node_id=current_node_id)
        join_with = current_node_id
    return current_node_id



import jsonpickle
import pickle
from typing import Dict, List, Union, Tuple

from pangtreebuild.affinity_tree import tree
from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.DAGMaf import DAGMaf


class TaskParameters:
    """Parameters provided for single execution of the PangtreeBuild."""

    def __init__(self,
                 running_time: str = None,
                 multialignment_file_path: str = None,
                 multialignment_format: str = None,
                 datatype: str = None,
                 metadata_file_path: str = None,
                 blosum_file_path: str = None,
                 output_path: str = None,
                 output_po: bool = None,
                 output_fasta: bool = None,
                 output_with_nodes: bool = None,
                 verbose: bool = None,

                 raw_maf: bool = None,
                 fasta_provider: str = None,
                 cache: bool = None,
                 missing_base_symbol: str = None,
                 fasta_source_file: str = None,

                 consensus_type: str = None,
                 hbmin: float = None,
                 stop: float = None,
                 p: float = None
                 ):
        self.running_time: str = running_time
        self.multialignment_file_path: str = str(multialignment_file_path)
        self.multialignment_format: str = str(multialignment_format)

        self.datatype: str = str(datatype)
        self.metadata_file_path: str = str(metadata_file_path)
        self.blosum_file_path: str = str(blosum_file_path)

        self.output_path: str = str(output_path)
        self.output_po: bool = output_po
        self.generate_fasta: bool = output_fasta
        self.output_with_nodes: bool = output_with_nodes
        self.verbose: bool = verbose

        self.raw_maf: bool = raw_maf
        self.fasta_complementation_option: str = fasta_provider
        self.cache: bool = cache
        self.missing_base_symbol: str = missing_base_symbol
        self.fasta_source_file: str = fasta_source_file

        self.consensus_type: str = consensus_type
        self.hbmin: float = hbmin

        self.stop: float = stop
        self.p: float = p


class Node:
    """Describes node of Poagraph data structure."""

    def __init__(self,
                 node_id: int,
                 base: str,
                 column_id: int,
                 block_id: int,
                 aligned_to: int):
        self.id: int = node_id
        self.base: str = base
        self.column_id: int = column_id
        self.block_id: int = block_id
        self.aligned_to: int = aligned_to


class Sequence:
    """Describes genome _sequences present in specific Poagraph."""

    def __init__(self,
                 sequence_int_id: int,
                 sequence_str_id: str,
                 metadata: Dict[str, Union[str, float]],
                 nodes_ids: List[int]):
        self.sequence_int_id: int = sequence_int_id
        self.sequence_str_id: str = sequence_str_id
        self.metadata: Dict[str, Union[str, float]] = metadata
        self.nodes_ids: List[List[int]] = nodes_ids


class AffinityNode:
    """Describes node present in Affinity Tree.

    It is connected with specific Poagraph structure as Poagraph nodes ids
    are listed here."""

    def __init__(self,
                 affinity_node_id: int,
                 name: str,
                 parent_node_id: int,
                 children_nodes_ids: List[int],
                 comp_to_all_sequences: Dict[str, float],
                 sequences_int_ids: List[int],
                 poagraph_nodes_ids: List[int],
                 mincomp: float):
        self.affinity_node_id: int = affinity_node_id
        self.name: str = name
        self.parent: int = parent_node_id
        self.children: List[int] = children_nodes_ids
        self.comp_to_all_sequences: Dict[str, float] = comp_to_all_sequences
        self.sequences_int_ids: List[int] = sequences_int_ids
        self.nodes_ids: List[int] = poagraph_nodes_ids
        self.mincomp: float = mincomp


class MafEdge:
    """Describes edge between two maf blocks."""

    def __init__(self,
                 edge_type: Tuple[int, int] = None,
                 sequences: List[str] = None,
                 to_block: int = None):
        self.edge_type: Tuple[int, int] = edge_type
        self.sequences: List[str] = sequences
        self.to_block: int = to_block


class MafNode:
    """Describes single block in maf file converted to DAGMaf."""

    def __init__(self,
                 node_id: int = None,
                 orient: int = None,
                 out_edges: List[MafEdge] = None
                 ):
        self.node_id: int = node_id
        self.orient: int = orient
        self.out_edges: List[MafEdge] = out_edges


class PangenomeJSON:
    """Describes data generated in single pang program call and parameters."""

    def __init__(self,
                 task_parameters: TaskParameters,
                 sequences: List[Sequence],
                 nodes: List[Node],
                 dagmaf_nodes: List[MafNode],
                 affinity_tree: List[AffinityNode]):
        self.task_parameters: TaskParameters = task_parameters
        self.sequences: List[Sequence] = sequences
        self.nodes: List[Node] = nodes
        self.dagmaf_nodes = dagmaf_nodes
        self.affinitytree = affinity_tree


def to_PangenomeJSON(task_parameters: TaskParameters = None,
                     poagraph: graph.Poagraph = None,
                     dagmaf: DAGMaf = None,
                     affinity_tree: tree.AffinityTree = None):
    """Converts pangenome specific objects to jsonable representation."""

    paths_seq_id_to_int_id = dict()
    pangenome_nodes = []
    pangenome_sequences = []
    dagmaf_nodes = []
    affinity_tree_nodes = []

    if poagraph:
        pangenome_nodes = [Node(node_id=node.node_id,
                                base=node.get_base(),
                                column_id=node.column_id,
                                block_id=node.block_id,
                                aligned_to=node.aligned_to)
                           for node in poagraph.nodes]

        sorted_sequences_ids = sorted(poagraph.sequences.keys())
        paths_seq_id_to_int_id = {seq_id: i
                                  for i, seq_id in enumerate(sorted_sequences_ids)}
        pangenome_sequences = [Sequence(sequence_int_id=paths_seq_id_to_int_id[seq_id],
                                        sequence_str_id=str(seq_id),
                                        metadata=poagraph.sequences[seq_id].seqmetadata,
                                        nodes_ids=[path
                                                   for path in poagraph.sequences[seq_id].paths]
                                                  if task_parameters.output_with_nodes
                                                  else []
                                        )
                               for seq_id in sorted_sequences_ids]

    # if dagmaf:
    #     dagmaf_nodes = [MafNode(id_=n.id_,
    #                             orient=n.orient,
    #                             out_edges=[MafEdge(edge_type=edge.edge_type,
    #                                                _sequences=edge._sequences,
    #                                                to_block=edge.to) for edge in n.out_edges])
    #                     for n in dagmaf.dagmaf_nodes]

    if affinity_tree:
        affinity_tree_nodes = [AffinityNode(affinity_node_id=consensus_node.id_,
                                            name=f"CONSENSUS{consensus_node.id_}",
                                            parent_node_id=consensus_node.parent,
                                            children_nodes_ids=consensus_node.children,
                                            comp_to_all_sequences={seq_id: comp.base_value().value
                                                                  for seq_id, comp
                                                                  in consensus_node.compatibilities.items()},
                                            sequences_int_ids=[paths_seq_id_to_int_id[seq_id]
                                                                  for seq_id
                                                                  in consensus_node.sequences
                                                                  if seq_id in paths_seq_id_to_int_id.keys()],
                                            # poagraph_nodes_ids=[],
                                            poagraph_nodes_ids=consensus_node.consensus
                                                if task_parameters.output_with_nodes else [],
                                            mincomp=consensus_node.mincomp.base_value().value)
                             for consensus_node in affinity_tree.nodes]
    return PangenomeJSON(task_parameters,
                         pangenome_sequences,
                         pangenome_nodes,
                         dagmaf_nodes,
                         affinity_tree_nodes
                         )


def to_json(pangenomejson: PangenomeJSON) -> str:
    # jsonpickle.set_encoder_options('simplejson', indent=4)
    return jsonpickle.encode(pangenomejson, unpicklable=False)


def to_pickle(pangenomejson: PangenomeJSON) -> str:
    """PangenomeJSON to str by pickling.
    """

    return pickle.dumps(pangenomejson)


def load_pickle(s: str) -> PangenomeJSON:
    """Str do PangenomeJSON by depickling.
    """

    return pickle.loads(s)


def str_to_PangenomeJSON(s: str) -> PangenomeJSON:
    """Str to PangenomeJSON.
    """

    pangenome_dict = jsonpickle.loads(s)
    if 'task_parameters' in pangenome_dict:
        task_parameters = TaskParameters()
        for k, v in pangenome_dict['task_parameters'].items():
            task_parameters.__dict__[k] = v
    if 'dagmaf_nodes' in pangenome_dict:
        dagmaf_nodes = [MafNode(node_id=dagmaf_node['node_id'],
                                orient=dagmaf_node['orient'],
                                out_edges=[MafEdge(edge_type=edge['edge_type'],
                                                   sequences=edge['sequences'],
                                                   to_block=edge['to_block']) for edge in dagmaf_node['out_edges']])
                        for dagmaf_node in pangenome_dict['dagmaf_nodes']]

    if "nodes" in pangenome_dict:
        pangenome_nodes = [Node(node_id=node["id"],
                                base=node['base'],
                                column_id=node['column_id'],
                                block_id=node['block_id'],
                                aligned_to=node['aligned_to'])
                           for node in pangenome_dict['nodes']]

    if 'sequences' in pangenome_dict:
        pangenome_sequences = [Sequence(sequence_int_id=sequence['sequence_int_id'],
                                        sequence_str_id=sequence['sequence_str_id'],
                                        metadata=sequence['metadata'],
                                        nodes_ids=sequence['nodes_ids']
                                        )
                               for sequence in pangenome_dict['sequences']]

    if 'affinitytree' in pangenome_dict:
        affinity_tree_nodes = [AffinityNode(affinity_node_id=consensus_node['affinity_node_id'],
                                          name=consensus_node['name'],
                                          parent_node_id=consensus_node['parent'],
                                          children_nodes_ids=consensus_node['children'],
                                          comp_to_all_sequences=consensus_node['comp_to_all_sequences'],
                                          sequences_int_ids=consensus_node['sequences_int_ids'],
                                          poagraph_nodes_ids=consensus_node['nodes_ids'],
                                          mincomp=consensus_node['mincomp'])
                              for consensus_node in pangenome_dict['affinitytree']]

    return PangenomeJSON(task_parameters,
                         pangenome_sequences,
                         pangenome_nodes,
                         dagmaf_nodes,
                         affinity_tree_nodes
                         )



# -*- coding: utf-8 -*-
import networkx as nx
import six
from .mafreader import read_maf
from .weighted_edges import weight, Edge

def _out_edges(v, blocks, ub, G):
    # Return a list of nodes connected to node v by edge leaving the vertex v   
    return [u for u in G[v] if ub >= blocks[u].order() > blocks[v].order()]

def _in_edges(v, blocks, lb, G):
    # Return a list of nodes connected to node v by edge coming in to the vertex v     
    return [u for u in G[v] if lb < blocks[u].order() < blocks[v].order()]  

def _dfs_f(root, ub, G, blocks):
    visited = set()
    stack = [root,]
    while stack:
        node = stack.pop()
        if blocks[node].order()==ub: return []
        if node not in visited:
            visited.add(node)
            stack.extend([x for x in _out_edges(node, blocks, ub, G) if x not in visited])
    return list(visited) 

def _dfs_b(root, lb, G, blocks):
    visited = set()
    stack = [root,]
    while stack:
        node = stack.pop()
        if node not in visited:
            visited.add(node)
            stack.extend([x for x in _in_edges(node, blocks, lb, G) if x not in visited])
    return list(visited)

def _reorder(R_f, R_b, blocks):
    R_f.sort(key = lambda x: blocks[x].order()) 
    R_b.sort(key = lambda x: blocks[x].order())
    L = R_b + R_f
    O = sorted(blocks[x].order() for x in L)
    if six.PY2:
        for i in xrange(len(L)):
            blocks[L[i]].reorder(O[i])
    else:
        for i in range(len(L)):
            blocks[L[i]].reorder(O[i])

def _add_edge_within_component(x, y, G, blocks):
    # Add edge between blocks of the same component
    lb = blocks[y].order()
    ub = blocks[x].order()
    if lb is ub: return
    elif lb < ub:
        R_f = _dfs_f(y, ub, G, blocks)
        if R_f:
            R_b = _dfs_b(x, lb, G, blocks)
            _reorder(R_f, R_b, blocks)
            G.add_edge(x,y)
    else: 
        G.add_edge(x,y)
       
def _add_edge_between_components(e, blocks):
    # Add edge between blocks of different components
    reverse, flank = 1, 1
    if blocks[e.left].size() < blocks[e.right].size():
        e = Edge(e.right, e.left, e.type[::-1])
    if blocks[e.left].orientation()*blocks[e.right].orientation() is e.type[0]*e.type[1]:
        reverse = -1
    if blocks[e.left].orientation()*e.type[0] < 0: 
        flank = -1
    blocks[e.right].unionto(blocks[e.left], reverse, flank)

def connect_components(blocks):
    if len(blocks) != blocks[0].size(): 
        d = {blocks[0].find(): 0}
        n = blocks[0].maximum()
        for block in blocks:
            if block.find() not in d:
                i = n - block.minimum() + 1
                d[block.find()] = i 
                block.reorder(i + block.order())
                n += block.size()
            else:
                block.reorder(d[block.find()] + block.order())
            block.orient_block()
    else:
        for block in blocks:
            block.orient_block()
            
def set_out_edges(d, blocks):
    for edge in d:
        edge_type = (blocks[edge.left].orientation()*edge.type[0], blocks[edge.right].orientation()*edge.type[1])
        for tup in d[edge][0]:
            tup[0].set_start_position(blocks[edge.left].alignment, blocks[edge.left].orientation())
            tup[1].set_start_position(blocks[edge.right].alignment, blocks[edge.right].orientation())   
        if blocks[edge.left].order() < blocks[edge.right].order():
            blocks[edge.left].add_out_edges(edge.right, edge_type, d[edge][0])
        elif blocks[edge.left].order() > blocks[edge.right].order():
            sequences = [x[::-1] for x in d[edge][0]]
            edge_type = edge_type[::-1]
            blocks[edge.right].add_out_edges(edge.left, edge_type, sequences)
        else:
            blocks[edge.left].add_out_edges(edge.right, edge.type, d[edge][0])

def sort_mafblocks(maf_file):   
    blocks, seq = read_maf(maf_file) # blocks - list of Block instances
    d = weight(seq)
    edges = sorted(d.keys(), key=lambda x: (d[x][1], x.type, x.left, x.right)) # list of edges sorted by the weight
    G = nx.Graph()
    for e in edges:
        if blocks[e.left].find() is blocks[e.right].find():
            if blocks[e.left].orientation()*blocks[e.right].orientation() is e.type[0]*e.type[1]:
                continue
            elif blocks[e.left].orientation()*e.type[0] > 0:
                _add_edge_within_component(e.left, e.right, G, blocks)
            else:
                _add_edge_within_component(e.right, e.left, G, blocks)           
        else:
            G.add_edge(e.left, e.right)
            _add_edge_between_components(e, blocks)
    set_out_edges(d, blocks)
    connect_components(blocks)
    blocks = sorted(blocks, key = lambda b: b.order())
    return blocks




import unittest
from pathlib import Path

from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.builders import maf2poagraph
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.tools import pathtools


def nid(x): return graph.NodeID(x)


def bid(x): return graph.BlockID(x)


class Maf2poagraphTests(unittest.TestCase):

    def setUp(self):
        metadata_path = Path(__file__).parent.joinpath("../seq_metadata.csv").resolve()
        self.metadatacsv = msa.MetadataCSV(pathtools.get_file_content_stringio(metadata_path), metadata_path)
        self.maf_files_dir = Path(__file__).parent.joinpath("maf_files").resolve()

    def test_1_messy_sequences(self):
        maf_path = self.maf_files_dir.joinpath("test_1_messy_sequences.maf")
        expected_nodes = [
            graph.Node(node_id=nid(0), base=graph.Base('A'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(1), base=graph.Base('A'), aligned_to=nid(2), block_id=bid(0)),
            graph.Node(node_id=nid(2), base=graph.Base('C'), aligned_to=nid(1), block_id=bid(0)),
            graph.Node(node_id=nid(3), base=graph.Base('T'), aligned_to=None, block_id=bid(0)),
            graph.Node(node_id=nid(4), base=graph.Base('C'), aligned_to=nid(5), block_id=bid(0)),
            graph.Node(node_id=nid(5), base=graph.Base('G'), aligned_to=nid(4), block_id=bid(0)),
            graph.Node(node_id=nid(6), base=graph.Base('A'), aligned_to=None, block_id=bid(1)),
            graph.Node(node_id=nid(7), base=graph.Base('C'), aligned_to=None, block_id=bid(1)),
            graph.Node(node_id=nid(8), base=graph.Base('G'), aligned_to=None, block_id=bid(1)),
            graph.Node(node_id=nid(9), base=graph.Base('C'), aligned_to=nid(10), block_id=bid(2)),
            graph.Node(node_id=nid(10), base=graph.Base('G'), aligned_to=nid(9), block_id=bid(2)),
            graph.Node(node_id=nid(11), base=graph.Base('T'), aligned_to=None, block_id=bid(2)),
            graph.Node(node_id=nid(12), base=graph.Base('C'), aligned_to=None, block_id=bid(2)),
            graph.Node(node_id=nid(13), base=graph.Base('C'), aligned_to=None, block_id=bid(2)),
            graph.Node(node_id=nid(14), base=graph.Base('A'), aligned_to=None, block_id=bid(2)),
        ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [1, 3, 4, 6, 8, 9, 11, 12])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [2, 3, 4, 10, 11, 12, 13, 14])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 2, 5, 6, 7, 10, 11, 12, 14])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'}))
        }
        actual_nodes, actual_sequences = maf2poagraph.get_poagraph(msa.Maf(pathtools.get_file_content_stringio(maf_path),
                                                                           maf_path),
                                                                   self.metadatacsv)

        self.assertEqual(expected_nodes, actual_nodes)
        self.assertEqual(expected_sequences, actual_sequences)


if __name__ == '__main__':
    unittest.main()



from pangtreebuild.mafgraph.sorter import sort_mafblocks
from pangtreebuild.pangenome import DAGMaf
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.tools import logprocess

global_logger = logprocess.get_global_logger()


def get_dagmaf(maf: msa.Maf) -> DAGMaf.DAGMaf:
    """Converts MAF to DagMaf.

    Args:
        maf: MAF to be converted.

    Returns:
         DagMaf built from the MAF.
    """
    sorted_blocks = sort_mafblocks(maf.filecontent)
    dagmafnodes = [
        DAGMaf.DAGMafNode(block_id=b.id,
                          alignment=b.alignment,
                          orient=b.orient,
                          order=b.order(),
                          out_edges=b.out_edges)
        for b in sorted_blocks
    ]
    return DAGMaf.DAGMaf(dagmafnodes)



import shutil
import unittest
from pathlib import Path

from ddt import ddt

from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import missings, msa
from pangtreebuild.tools import pathtools


@ddt
class FromNCBI_FastaDiskCache_Tests(unittest.TestCase):
    def setUp(self) -> None:
        self.fasta_provider = missings.FromNCBI(use_cache=False)

    # @unittest.skip('Internet connection required -> long execution')
    def test_1_download_sequence_and_save_to_cache(self):
        cache_dir_path = pathtools.get_child_path(Path.cwd(), ".fastacache")
        if cache_dir_path.exists():
            shutil.rmtree(cache_dir_path)

        ncbi_fasta_provider = missings.FromNCBI(use_cache=True)
        sequence_id = msa.SequenceID("AB050936v1")

        _ = ncbi_fasta_provider.get_base(sequence_id, 0)

        # cache directory creation
        cache_directory_created = cache_dir_path.exists()
        self.assertTrue(cache_directory_created)

        # file creation
        files_in_cache_dircetory = [*cache_dir_path.glob("*")]
        expected_filepath = pathtools.get_child_path(cache_dir_path, f"{sequence_id}.fasta")
        file_created_in_cache = expected_filepath in files_in_cache_dircetory
        self.assertTrue(file_created_in_cache)

        # file content
        control_fasta_path = Path(__file__).parent.joinpath('fasta_ncbi/AB050936.1.fasta').resolve()

        with open(control_fasta_path) as fasta_file_hanlder:
            expected_content = fasta_file_hanlder.read()
        with open(expected_filepath) as fasta_file_handler:
            actual_content = fasta_file_handler.read()
        self.assertEqual(expected_content, actual_content)

    def test_2_read_seqeunce_from_cache_instead_downloading(self):
        fasta_provider = missings.FromNCBI(use_cache=True)
        cache_dir_path = pathtools.get_child_path(Path.cwd(), ".fastacache")
        if cache_dir_path.exists():
            shutil.rmtree(cache_dir_path)

        cache_dir_path.mkdir()
        sequence_id = msa.SequenceID("seq1")
        fake_sequence = "foo"
        expected_base = graph.Base("o")
        fake_fasta_path = pathtools.get_child_path(cache_dir_path, f"{sequence_id}.fasta")
        with open(fake_fasta_path, 'w') as fake_fasta_handler:
            fake_fasta_handler.write(f">{sequence_id} cached\n{fake_sequence}")

        actual_base = fasta_provider.get_base(sequence_id, 2)

        self.assertEqual(expected_base, actual_base)


if __name__ == '__main__':
    unittest.main()



import logging
import logging.config
import os
import sys
from pathlib import Path

from pangtreebuild.tools import pathtools

logging_config_path = Path(os.path.abspath(__file__)).joinpath('../../logging.conf').resolve()
logging.config.fileConfig(logging_config_path)


def get_global_logger() -> logging.Logger:
    """Returns global logger."""

    return logging.getLogger("")


def get_logger(logger_name: str) -> logging.Logger:
    """Returns logger of given logger_name.

    Args:
        logger_name: Name of the logger to return.

    Returns:
        Logger with name logger_name.
    """

    return logging.getLogger(logger_name)


def add_file_handler_to_logger(outputdir: Path,
                               logger_name: str,
                               filename: str,
                               handlerformat: str = "%(levelname)s - %(message)s",
                               propagate: bool = True) -> None:
    """Adds new file handle to given logger.

    Args:
        outputdir: Directory where the file will be stored.
        logger_name: Logger to modify.
        filename: New logger name.
        handlerformat: Log message format.
        propagate: Whether logger should propagate.
    """

    logger = logging.getLogger(logger_name)
    fh = logging.FileHandler(pathtools.get_child_path(outputdir, filename))
    ft = logging.Formatter(handlerformat, datefmt="%x-%X")
    fh.setFormatter(ft)
    logger.propagate = propagate
    logger.addHandler(fh)


def add_console_handler_to_logger(logger_name: str,
                                  propagate: bool):
    """Adds new console handle to given logger.

    Args:
        logger_name: Logger to modify.
        propagate: Whether logger should propagate.
    """

    console_logger = logging.getLogger(logger_name)
    console_logger.propagate = propagate
    console_handler = logging.StreamHandler(sys.stdout)
    console_logger.addHandler(console_handler)


def disable_all_loggers() -> None:
    """Turns off all loggers."""

    logging.root.disabled = True


def remove_console_handler_from_root_logger() -> None:
    """Turn off console logger completely."""

    logger = logging.getLogger("root").parent
    for lh in logger.handlers[:]:
        if type(lh) == logging.StreamHandler and lh.stream.name == "<stdout>":
            logger.removeHandler(lh)



import unittest
from pathlib import Path

from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.builders import po2poagraph
from pangtreebuild.pangenome.parameters import msa
from pangtreebuild.tools import pathtools


def nid(x): return graph.NodeID(x)


def bid(x): return graph.Base(x)


class Po2poagraphTests(unittest.TestCase):

    def setUp(self):
        metadata_path = Path(__file__).parent.joinpath("../seq_metadata.csv").resolve()
        self.metadatacsv = msa.MetadataCSV(pathtools.get_file_content_stringio(metadata_path), metadata_path)
        self.po_files_dir = Path(__file__).parent.joinpath("po_files").resolve()
        
    def test_1_typical_poagraph(self):
        po_path = self.po_files_dir.joinpath("test_1.po")

        expected_nodes = [graph.Node(node_id=nid(0), base=bid('A'), aligned_to=nid(1)),
                          graph.Node(node_id=nid(1), base=bid('G'), aligned_to=nid(0)),
                          graph.Node(node_id=nid(2), base=bid('C'), aligned_to=nid(3)),
                          graph.Node(node_id=nid(3), base=bid('G'), aligned_to=nid(2)),
                          graph.Node(node_id=nid(4), base=bid('A'), aligned_to=nid(5)),
                          graph.Node(node_id=nid(5), base=bid('T'), aligned_to=nid(4)),
                          graph.Node(node_id=nid(6), base=bid('G'), aligned_to=None),
                          graph.Node(node_id=nid(7), base=bid('G'), aligned_to=None),
                          graph.Node(node_id=nid(8), base=bid('A'), aligned_to=nid(9)),
                          graph.Node(node_id=nid(9), base=bid('C'), aligned_to=nid(10)),
                          graph.Node(node_id=nid(10), base=bid('G'), aligned_to=nid(11)),
                          graph.Node(node_id=nid(11), base=bid('T'), aligned_to=nid(8)),
                          graph.Node(node_id=nid(12), base=bid('A'), aligned_to=nid(13)),
                          graph.Node(node_id=nid(13), base=bid('C'), aligned_to=nid(12)),
                          graph.Node(node_id=nid(14), base=bid('T'), aligned_to=None),
                          graph.Node(node_id=nid(15), base=bid('A'), aligned_to=nid(16)),
                          graph.Node(node_id=nid(16), base=bid('C'), aligned_to=nid(17)),
                          graph.Node(node_id=nid(17), base=bid('G'), aligned_to=nid(15))
                          ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0, 2, 4, 6, 7, 8, 12, 14, 16])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [1, 2, 5, 6, 7, 9])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [3, 4, 6, 7, 10, 12, 14, 17])])],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [graph.SeqPath([*map(nid, [11, 13, 14, 15])])],
                               graph.SequenceMetadata({'group': '2'}))
        }

        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        nodes, sequences = po2poagraph.get_poagraph(msa.Po(pathtools.get_file_content_stringio(po_path), po_path),
                                                    self.metadatacsv)
        actual_poagraph = graph.Poagraph(nodes, sequences)
        self.assertEqual(expected_poagraph, actual_poagraph)

    def test_2_consensuses_and_empty_sequences(self):
        po_path = self.po_files_dir.joinpath("test_2.po")

        expected_nodes = [graph.Node(node_id=nid(0), base=bid('C'), aligned_to=nid(1)),
                          graph.Node(node_id=nid(1), base=bid('T'), aligned_to=nid(0)),
                          graph.Node(node_id=nid(2), base=bid('A'), aligned_to=nid(3)),
                          graph.Node(node_id=nid(3), base=bid('G'), aligned_to=nid(2)),
                          graph.Node(node_id=nid(4), base=bid('C'), aligned_to=None),
                          graph.Node(node_id=nid(5), base=bid('T'), aligned_to=None),
                          graph.Node(node_id=nid(6), base=bid('A'), aligned_to=nid(7)),
                          graph.Node(node_id=nid(7), base=bid('T'), aligned_to=nid(6)),
                          graph.Node(node_id=nid(8), base=bid('G'), aligned_to=None)
                          ]

        expected_sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0, 3, 4, 5, 6, 8])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [1, 2, 4, 5, 7, 8])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [],
                               graph.SequenceMetadata({'group': '2'})),
            msa.SequenceID('CONSENS0'):
                graph.Sequence(msa.SequenceID('CONSENS0'),
                               [graph.SeqPath([*map(nid, [0, 3, 4, 5, 7, 8])])],
                               graph.SequenceMetadata({})),
            msa.SequenceID('CONSENS1'):
                graph.Sequence(msa.SequenceID('CONSENS1'),
                               [graph.SeqPath([*map(nid, [1, 2, 4, 5, 6, 8])])],
                               graph.SequenceMetadata({}))
        }

        expected_poagraph = graph.Poagraph(expected_nodes, expected_sequences)
        nodes, sequences = po2poagraph.get_poagraph(msa.Po(pathtools.get_file_content_stringio(po_path), po_path),
                                                    self.metadatacsv)
        actual_poagraph = graph.Poagraph(nodes, sequences)
        self.assertEqual(expected_poagraph, actual_poagraph)


if __name__ == '__main__':
    unittest.main()



import unittest
from ddt import ddt

from pangtreebuild.affinity_tree import poa
from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import msa


def nid(x): return graph.NodeID(x)


def b(x): return graph.Base(x)


@ddt
class PoagraphPOTranslator_read_top_consensus_Test(unittest.TestCase):

    def setUp(self):
        nodes = [graph.Node(node_id=nid(0), base=b('T'), aligned_to=None, ),
                 graph.Node(node_id=nid(1), base=b('A'), aligned_to=nid(2)),
                 graph.Node(node_id=nid(2), base=b('G'), aligned_to=nid(1)),
                 graph.Node(node_id=nid(3), base=b('A'), aligned_to=nid(4)),
                 graph.Node(node_id=nid(4), base=b('C'), aligned_to=nid(3)),
                 graph.Node(node_id=nid(5), base=b('A'), aligned_to=nid(6)),
                 graph.Node(node_id=nid(6), base=b('C'), aligned_to=nid(7)),
                 graph.Node(node_id=nid(7), base=b('G'), aligned_to=nid(8)),
                 graph.Node(node_id=nid(8), base=b('T'), aligned_to=nid(5)),
                 graph.Node(node_id=nid(9), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(10), base=b('C'), aligned_to=nid(11)),
                 graph.Node(node_id=nid(11), base=b('T'), aligned_to=nid(10)),
                 graph.Node(node_id=nid(12), base=b('G'), aligned_to=None),
                 graph.Node(node_id=nid(13), base=b('A'), aligned_to=nid(14)),
                 graph.Node(node_id=nid(14), base=b('C'), aligned_to=nid(13))
                 ]

        sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0, 1, 3, 5, 9, 10, 13])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [1, 3, 6, 9, 11])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [2, 4, 7, 9, 11, 12])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [graph.SeqPath([*map(nid, [2, 4, 8, 9, 11, 12, 14])])],
                               graph.SequenceMetadata({'group': '1'})),
        }

        self.poagraph = graph.Poagraph(nodes, sequences)

    def test_read_consensus_path_seq1_only_in_input(self):
        translator = poa._PoagraphPOTranslator(self.poagraph,
                                               [msa.SequenceID('seq1')])
        _ = translator.get_input_po_content()

        poa_lines = ["VERSION=pangenome\n",
                     "NAME=pangenome\n",
                     "TITLE=pangenome\n",
                     "LENGTH=5\n",
                     "SOURCECOUNT=2\n",
                     "SOURCENAME=seq1\n",
                     "SOURCEINFO=5 0 100 0 seq1\n",
                     "SOURCENAME=CONSENS0\n",
                     "SOURCEINFO=5 0 100 0 CONSENS0\n",
                     "a:S0S1\n",
                     "a:L0S0S1\n",
                     "c:L1S0S1\n",
                     "a:L2S0S1\n",
                     "t:L2S0S1"]
        actual_consensus_path = translator.read_consensus_paths(poa_lines, [0])
        expected_consensus_path = [1, 3, 6, 9, 11]
        self.assertEqual(expected_consensus_path, actual_consensus_path[0].path)

    def test_subpoagraph_construction_from_poagraph_keep_seq_0_1(self):
        translator = poa._PoagraphPOTranslator(self.poagraph,
                                               [msa.SequenceID('seq0'),
                                                msa.SequenceID('seq1')])
        actual_po_content = translator.get_input_po_content()
        expected_po_content = "VERSION=pangenome\n"\
                              "NAME=pangenome\n"\
                              "TITLE=pangenome\n"\
                              "LENGTH=9\n"\
                              "SOURCECOUNT=2\n"\
                              "SOURCENAME=seq0\n"\
                              "SOURCEINFO=7 0 0 -1 seq0\n"\
                              "SOURCENAME=seq1\n"\
                              "SOURCEINFO=5 1 100 -1 seq1\n"\
                              "t:S0\n"\
                              "a:L0S0S1\n"\
                              "a:L1S0S1\n"\
                              "a:L2S0A4\n"\
                              "c:L2S1A3\n"\
                              "a:L3L4S0S1\n"\
                              "c:L5S0A7\n"\
                              "t:L5S1A6\n"\
                              "a:L6S0"
        self.assertEqual(expected_po_content, actual_po_content)

    def test_subpoagraph_construction_from_poagraph_keep_seq3(self):
        translator = poa._PoagraphPOTranslator(self.poagraph,
                                               [msa.SequenceID('seq3')])
        actual_po_content = translator.get_input_po_content()
        expected_po_content = "VERSION=pangenome\n" \
                              "NAME=pangenome\n" \
                              "TITLE=pangenome\n" \
                              "LENGTH=7\n" \
                              "SOURCECOUNT=1\n" \
                              "SOURCENAME=seq3\n" \
                              "SOURCEINFO=7 0 100 -1 seq3\n" \
                              "g:S0\n" \
                              "c:L0S0\n" \
                              "t:L1S0\n" \
                              "a:L2S0\n" \
                              "t:L3S0\n" \
                              "g:L4S0\n" \
                              "c:L5S0"
        self.assertEqual(expected_po_content, actual_po_content)

    def test_subpoagraph_construction_full_graph(self):
        nodes = [graph.Node(node_id=nid(0), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(1), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(2), base=b('C'), aligned_to=None),
                 graph.Node(node_id=nid(3), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(4), base=b('T'), aligned_to=None)
                 ]

        sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [0, 1, 2, 3, 4])])],
                               graph.SequenceMetadata({'group': '1'}))
        }
        poagraph = graph.Poagraph(nodes, sequences)
        translator = poa._PoagraphPOTranslator(poagraph, [msa.SequenceID('seq0')])
        actual_po_content = translator.get_input_po_content()
        expected_po_content = "VERSION=pangenome\n" \
                              "NAME=pangenome\n" \
                              "TITLE=pangenome\n" \
                              "LENGTH=5\n" \
                              "SOURCECOUNT=1\n" \
                              "SOURCENAME=seq0\n" \
                              "SOURCEINFO=5 0 100 -1 seq0\n" \
                              "a:S0\n" \
                              "a:L0S0\n" \
                              "c:L1S0\n" \
                              "a:L2S0\n" \
                              "t:L3S0"
        self.assertEqual(expected_po_content, actual_po_content)

    def test_subpoagraph_should_omit_edges_1(self):
        nodes = [graph.Node(node_id=nid(0), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(1), base=b('C'), aligned_to=None),
                 graph.Node(node_id=nid(2), base=b('C'), aligned_to=None)]

        sequences = {
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 2])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 1, 2])])],
                               graph.SequenceMetadata({'group': '1'}))
        }
        poagraph = graph.Poagraph(nodes, sequences)

        translator = poa._PoagraphPOTranslator(poagraph, [msa.SequenceID('seq2')])
        actual_po_content = translator.get_input_po_content()
        expected_po_content = "VERSION=pangenome\n" \
                              "NAME=pangenome\n" \
                              "TITLE=pangenome\n" \
                              "LENGTH=3\n" \
                              "SOURCECOUNT=1\n" \
                              "SOURCENAME=seq2\n" \
                              "SOURCEINFO=3 0 100 -1 seq2\n" \
                              "a:S0\n" \
                              "c:L0S0\n" \
                              "c:L1S0"
        self.assertEqual(expected_po_content, actual_po_content)

    def test_subpoagraph_should_omit_edges_2(self):
        nodes = [graph.Node(node_id=nid(0), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(1), base=b('C'), aligned_to=None),
                 graph.Node(node_id=nid(2), base=b('C'), aligned_to=None)]

        sequences = {
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 2])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 1, 2])])],
                               graph.SequenceMetadata({'group': '1'}))
        }
        poagraph = graph.Poagraph(nodes, sequences)

        translator = poa._PoagraphPOTranslator(poagraph, [msa.SequenceID('seq1')])
        actual_po_content = translator.get_input_po_content()
        expected_po_content = "VERSION=pangenome\n" \
                              "NAME=pangenome\n" \
                              "TITLE=pangenome\n" \
                              "LENGTH=2\n" \
                              "SOURCECOUNT=1\n" \
                              "SOURCENAME=seq1\n" \
                              "SOURCEINFO=2 0 100 -1 seq1\n" \
                              "a:S0\n" \
                              "c:L0S0"

        self.assertEqual(expected_po_content, actual_po_content)

    def test_subpoagraph_should_omit_in_nodes_and_aligned_nodes(self):
        # original poagraph
        nodes = [graph.Node(node_id=nid(0), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(1), base=b('C'), aligned_to=nid(2)),
                 graph.Node(node_id=nid(2), base=b('T'), aligned_to=nid(1)),
                 graph.Node(node_id=nid(3), base=b('G'), aligned_to=None)]

        sequences = {
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 1, 3])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [0, 2, 3])])],
                               graph.SequenceMetadata({'group': '1'}))
        }
        poagraph = graph.Poagraph(nodes, sequences)

        translator = poa._PoagraphPOTranslator(poagraph, [msa.SequenceID('seq2')])
        actual_po_content = translator.get_input_po_content()
        expected_po_content = "VERSION=pangenome\n" \
                              "NAME=pangenome\n" \
                              "TITLE=pangenome\n" \
                              "LENGTH=3\n" \
                              "SOURCECOUNT=1\n" \
                              "SOURCENAME=seq2\n" \
                              "SOURCEINFO=3 0 100 -1 seq2\n" \
                              "a:S0\n" \
                              "t:L0S0\n" \
                              "g:L1S0"

        self.assertEqual(expected_po_content, actual_po_content)

    def test_subpoagraph_unfilled_nodes(self):
        symbol_for_uknown = '?'
        nodes = [graph.Node(node_id=nid(0), base=b('A'), aligned_to=nid(1)),
                 graph.Node(node_id=nid(1), base=b('C'), aligned_to=nid(0)),
                 graph.Node(node_id=nid(2), base=b('G'), aligned_to=None),
                 graph.Node(node_id=nid(3), base=b(symbol_for_uknown), aligned_to=None),
                 graph.Node(node_id=nid(4), base=b(symbol_for_uknown), aligned_to=None),
                 graph.Node(node_id=nid(5), base=b('G'), aligned_to=None),
                 graph.Node(node_id=nid(6), base=b('C'), aligned_to=None),
                 graph.Node(node_id=nid(7), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(5), base=b('T'), aligned_to=None)]

        sequences = {
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [0, 2, 3, 4, 7, 8])])],
                               graph.SequenceMetadata({'group': '1'})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [1, 2, 5, 6, 7, 8])])],
                               graph.SequenceMetadata({'group': '1'}))
        }
        poagraph = graph.Poagraph(nodes, sequences)

        translator = poa._PoagraphPOTranslator(poagraph, [msa.SequenceID('seq1'),
                                                          msa.SequenceID('seq2')])
        actual_po_content = translator.get_input_po_content()
        expected_po_content = "VERSION=pangenome\n" \
                              "NAME=pangenome\n" \
                              "TITLE=pangenome\n" \
                              "LENGTH=9\n" \
                              "SOURCECOUNT=2\n" \
                              "SOURCENAME=seq1\n" \
                              "SOURCEINFO=6 0 100 -1 seq1\n" \
                              "SOURCENAME=seq2\n" \
                              "SOURCEINFO=6 1 100 -1 seq2\n" \
                              "a:S0A1\n" \
                              "c:S1A0\n" \
                              "g:L0L1S0S1\n" \
                              f"{symbol_for_uknown}:L2S0\n" \
                              f"{symbol_for_uknown}:L3S0\n" \
                              "g:L2S1\n" \
                              "c:L5S1\n" \
                              "a:L4L6S0S1\n" \
                              "t:L7S0S1"
        self.assertEqual(expected_po_content, actual_po_content)
